# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from os_win._i18n import _
from os_win import constants
from os_win import exceptions
from os_win.utils import baseutils

LOG = logging.getLogger(__name__)


class DNSUtils(baseutils.BaseUtils):

    _DNS_NAMESPACE = '//%s/root/MicrosoftDNS'

    def __init__(self, host='.'):
        self._dns_manager_attr = None
        self._host = host

    @property
    def _dns_manager(self):
        if not self._dns_manager_attr:
            try:
                namespace = self._DNS_NAMESPACE % self._host
                self._dns_manager_attr = self._get_wmi_obj(namespace)
            except Exception:
                raise exceptions.DNSException(
                    _("Namespace %(namespace)s not found. Make sure "
                      "DNS Server feature is installed.") %
                    {'namespace': namespace})

        return self._dns_manager_attr

    def _get_zone(self, zone_name, ignore_missing=True):
        zones = self._dns_manager.MicrosoftDNS_Zone(Name=zone_name)
        if zones:
            return zones[0]
        if not ignore_missing:
            raise exceptions.DNSZoneNotFound(zone_name=zone_name)

    def zone_list(self):
        """Returns the current list of DNS Zones.

        """

        zones = self._dns_manager.MicrosoftDNS_Zone()
        return [x.Name for x in zones]

    def zone_exists(self, zone_name):
        return self._get_zone(zone_name) is not None

    def get_zone_properties(self, zone_name):
        zone = self._get_zone(zone_name, ignore_missing=False)

        zone_properties = {}
        zone_properties['zone_type'] = zone.ZoneType
        zone_properties['ds_integrated'] = zone.DsIntegrated
        zone_properties['data_file_name'] = zone.DataFile
        zone_properties['master_servers'] = zone.MasterServers or []

        return zone_properties

    def zone_create(self, zone_name, zone_type, ds_integrated,
                    data_file_name=None, ip_addrs=None,
                    admin_email_name=None):
        """Creates a DNS Zone and returns the path to the associated object.

        :param zone_name: string representing the name of the zone.
        :param zone_type: type of zone
            0 = Primary zone
            1 = Secondary zone, MUST include at least one master IP
            2 = Stub zone, MUST include at least one master IP
            3 = Zone forwarder, MUST include at least one master IP
        :param ds_integrated: Only Primary zones can be stored in AD
            True = the zone data is stored in the Active Directory
            False = the data zone is stored in files
        :param data_file_name(Optional): name of the data file associated
            with the zone.
        :param ip_addrs(Optional): IP addresses of the master DNS servers
            for this zone. Parameter type MUST be list
        :param admin_email_name(Optional): email address of the administrator
            responsible for the zone.
        """
        LOG.debug("Creating DNS Zone '%s'" % zone_name)
        if self.zone_exists(zone_name):
            raise exceptions.DNSZoneAlreadyExists(zone_name=zone_name)

        dns_zone_manager = self._dns_manager.MicrosoftDNS_Zone
        (zone_path,) = dns_zone_manager.CreateZone(
            ZoneName=zone_name,
            ZoneType=zone_type,
            DsIntegrated=ds_integrated,
            DataFileName=data_file_name,
            IpAddr=ip_addrs,
            AdminEmailname=admin_email_name)
        return zone_path

    def zone_delete(self, zone_name):
        LOG.debug("Deleting DNS Zone '%s'" % zone_name)

        zone_to_be_deleted = self._get_zone(zone_name)
        if zone_to_be_deleted:
            zone_to_be_deleted.Delete_()

    def zone_modify(self, zone_name, allow_update=None, disable_wins=None,
                    notify=None, reverse=None, secure_secondaries=None):
        """Modifies properties of an existing zone. If any parameter is None,

        then that parameter will be skipped and will not be taken into
        consideration.

        :param zone_name: string representing the name of the zone.
        :param allow_update:
            0 = No updates allowed.
            1 = Zone accepts both secure and nonsecure updates.
            2 = Zone accepts secure updates only.
        :param disable_wins: Indicates whether the WINS record is replicated.
            If set to TRUE, WINS record replication is disabled.
        :param notify:
            0 = Do not notify secondaries
            1 = Notify Servers listed on the Name Servers Tab
            2 = Notify the specified servers
        :param reverse: Indicates whether the Zone is reverse (TRUE)
            or forward (FALSE).
        :param secure_secondaries:
            0 = Allowed to Any host
            1 = Only to the Servers listed on the Name Servers tab
            2 = To the following servers (destination servers IP addresses
                are specified in SecondaryServers value)
            3 = Zone transfers not allowed
        """

        zone = self._get_zone(zone_name, ignore_missing=False)

        if allow_update is not None:
            zone.AllowUpdate = allow_update
        if disable_wins is not None:
            zone.DisableWINSRecordReplication = disable_wins
        if notify is not None:
            zone.Notify = notify
        if reverse is not None:
            zone.Reverse = reverse
        if secure_secondaries is not None:
            zone.SecureSecondaries = secure_secondaries

        zone.put()

    def zone_update(self, zone_name):
        LOG.debug("Updating DNS Zone '%s'" % zone_name)

        zone = self._get_zone(zone_name, ignore_missing=False)
        if (zone.DsIntegrated and
                zone.ZoneType == constants.DNS_ZONE_TYPE_PRIMARY):
            zone.UpdateFromDS()
        elif zone.ZoneType in [constants.DNS_ZONE_TYPE_SECONDARY,
                               constants.DNS_ZONE_TYPE_STUB]:
            zone.ForceRefresh()
        elif zone.ZoneType in [constants.DNS_ZONE_TYPE_PRIMARY,
                               constants.DNS_ZONE_TYPE_FORWARD]:
            zone.ReloadZone()

    def get_zone_serial(self, zone_name):
        # Performing a manual check to make sure the zone exists before
        # trying to retrieve the MicrosoftDNS_SOAType object. Otherwise,
        # the query for MicrosoftDNS_SOAType will fail with "Generic Failure"
        if not self.zone_exists(zone_name):
            # Return None if zone was not found
            return None

        zone_soatype = self._dns_manager.MicrosoftDNS_SOAType(
            ContainerName=zone_name)
        if not zone_soatype:
            return None
        # Serial number of the SOA record
        SOA = zone_soatype[0].SerialNumber
        return int(SOA)



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes

import os_win.conf
from os_win.utils.winapi import wintypes

CONF = os_win.conf.CONF

lib_handle = None

HBA_STATUS = ctypes.c_uint32
HBA_HANDLE = ctypes.c_uint32
HBA_PortType = ctypes.c_uint32
HBA_PortSpeed = ctypes.c_uint32
HBA_PortState = ctypes.c_uint32
HBA_COS = ctypes.c_uint32
HBA_FC4Types = ctypes.c_uint32 * 32
HBA_FCPBindingType = wintypes.INT


class HBA_WWN(ctypes.Structure):
    _fields_ = [('wwn', ctypes.c_ubyte * 8)]


class HBA_PortAttributes(ctypes.Structure):
    _fields_ = [('NodeWWN', HBA_WWN),
                ('PortWWN', HBA_WWN),
                ('PortFcId', ctypes.c_uint32),
                ('PortType', HBA_PortType),
                ('PortState', HBA_PortState),
                ('PortSupportedClassofService', HBA_COS),
                ('PortSupportedFc4Types', HBA_FC4Types),
                ('PortSymbolicName', wintypes.CHAR * 256),
                ('OSDeviceName', wintypes.CHAR * 256),
                ('PortSupportedSpeed', HBA_PortSpeed),
                ('PortSpeed', HBA_PortSpeed),
                ('PortMaxFrameSize', ctypes.c_uint32),
                ('FabricName', HBA_WWN),
                ('NumberOfDiscoveredPorts', ctypes.c_uint32)]


class HBA_FCPId(ctypes.Structure):
    _fields_ = [('FcId', ctypes.c_uint32),
                ('NodeWWN', HBA_WWN),
                ('PortWWN', HBA_WWN),
                ('FcpLun', ctypes.c_uint64)]


class HBA_ScsiId(ctypes.Structure):
    _fields_ = [('OSDeviceName', wintypes.CHAR * 256),
                ('ScsiBusNumber', ctypes.c_uint32),
                ('ScsiTargetNumber', ctypes.c_uint32),
                ('ScsiOSLun', ctypes.c_uint32)]


class HBA_FCPScsiEntry(ctypes.Structure):
    _fields_ = [('ScsiId', HBA_ScsiId),
                ('FcpId', HBA_FCPId)]


def get_target_mapping_struct(entry_count=0):
    class HBA_FCPTargetMapping(ctypes.Structure):
        _fields_ = [('NumberOfEntries', ctypes.c_uint32),
                    ('Entries', HBA_FCPScsiEntry * entry_count)]

        def __init__(self, entry_count):
            self.NumberOfEntries = entry_count
            self.Entries = (HBA_FCPScsiEntry * entry_count)()

    return HBA_FCPTargetMapping(entry_count)


class HBA_AdapterAttributes(ctypes.Structure):
    _fields_ = [('Manufacturer', wintypes.CHAR * 64),
                ('SerialNumber', wintypes.CHAR * 64),
                ('Model', wintypes.CHAR * 256),
                ('ModelDescription', wintypes.CHAR * 256),
                ('NodeWWN', HBA_WWN),
                ('NodeSymbolicName', wintypes.CHAR * 256),
                ('HardwareVersion', wintypes.CHAR * 256),
                ('DriverVersion', wintypes.CHAR * 256),
                ('OptionROMVersion', wintypes.CHAR * 256),
                ('FirmwareVersion', wintypes.CHAR * 256),
                ('VendorSpecificID', ctypes.c_uint32),
                ('NumberOfPorts', ctypes.c_uint32),
                ('DriverName', wintypes.CHAR * 256)]


def register():
    global lib_handle
    lib_handle = ctypes.cdll.LoadLibrary(CONF.os_win.hbaapi_lib_path)

    lib_handle.HBA_CloseAdapter.argtypes = [HBA_HANDLE]
    lib_handle.HBA_CloseAdapter.restype = None

    lib_handle.HBA_GetAdapterAttributes.argtypes = [
        HBA_HANDLE,
        ctypes.POINTER(HBA_AdapterAttributes)]
    lib_handle.HBA_GetAdapterAttributes.restype = HBA_STATUS

    lib_handle.HBA_GetAdapterName.argtypes = [
        ctypes.c_uint32,
        wintypes.PCHAR
    ]
    lib_handle.HBA_GetAdapterName.restype = HBA_STATUS

    lib_handle.HBA_GetAdapterPortAttributes.argtypes = [
        HBA_HANDLE,
        ctypes.c_uint32,
        ctypes.POINTER(HBA_PortAttributes)
    ]
    lib_handle.HBA_GetAdapterPortAttributes.restype = HBA_STATUS

    lib_handle.HBA_GetFcpTargetMapping.argtypes = [
        HBA_HANDLE,
        wintypes.PVOID
    ]
    lib_handle.HBA_GetFcpTargetMapping.restype = HBA_STATUS

    lib_handle.HBA_GetNumberOfAdapters.argtypes = []
    lib_handle.HBA_GetNumberOfAdapters.restype = ctypes.c_uint32

    lib_handle.HBA_OpenAdapter.argtypes = [wintypes.PCHAR]
    lib_handle.HBA_OpenAdapter.restype = HBA_HANDLE

    lib_handle.HBA_OpenAdapterByWWN.argtypes = [
        ctypes.POINTER(HBA_HANDLE),
        HBA_WWN]
    lib_handle.HBA_OpenAdapterByWWN.restype = HBA_STATUS

    lib_handle.HBA_ScsiInquiryV2.argtypes = [
        HBA_HANDLE,
        HBA_WWN,
        HBA_WWN,
        ctypes.c_uint64,
        ctypes.c_uint8,
        ctypes.c_uint8,
        wintypes.PVOID,
        ctypes.POINTER(ctypes.c_uint32),
        ctypes.POINTER(ctypes.c_uint8),
        wintypes.PVOID,
        ctypes.POINTER(ctypes.c_uint32)
    ]
    lib_handle.HBA_ScsiInquiryV2.restype = HBA_STATUS

    lib_handle.HBA_RefreshAdapterConfiguration.argtypes = []
    lib_handle.HBA_RefreshAdapterConfiguration.restype = None



# Copyright 2017 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock

from os_win.tests.unit import test_base
from os_win.utils import processutils
from os_win.utils.winapi import constants as w_const


@ddt.ddt
class ProcessUtilsTestCase(test_base.OsWinBaseTestCase):

    _autospec_classes = [
        processutils.win32utils.Win32Utils,
    ]

    def setUp(self):
        super(ProcessUtilsTestCase, self).setUp()
        self._setup_lib_mocks()

        self._procutils = processutils.ProcessUtils()
        self._win32_utils = self._procutils._win32_utils
        self._mock_run = self._win32_utils.run_and_check_output

        self.addCleanup(mock.patch.stopall)

    def _setup_lib_mocks(self):
        self._ctypes = mock.Mock()
        # This is used in order to easily make assertions on the variables
        # passed by reference.
        self._ctypes.byref = lambda x: (x, "byref")
        self._ctypes.c_wchar_p = lambda x: (x, 'c_wchar_p')
        self._ctypes.sizeof = lambda x: (x, 'sizeof')

        self._ctypes_patcher = mock.patch.multiple(
            processutils, ctypes=self._ctypes)
        self._ctypes_patcher.start()

        self._mock_kernel32 = mock.Mock()

        mock.patch.multiple(processutils,
                            kernel32=self._mock_kernel32).start()

    def test_create_job_object(self):
        job_handle = self._procutils.create_job_object(mock.sentinel.name)

        self._mock_run.assert_called_once_with(
            self._mock_kernel32.CreateJobObjectW,
            None,
            self._ctypes.c_wchar_p(mock.sentinel.name),
            error_ret_vals=[None],
            kernel32_lib_func=True)
        self.assertEqual(self._mock_run.return_value, job_handle)

    def test_set_information_job_object(self):
        self._procutils.set_information_job_object(
            mock.sentinel.job_handle,
            mock.sentinel.job_info_class,
            mock.sentinel.job_info)

        self._mock_run.assert_called_once_with(
            self._mock_kernel32.SetInformationJobObject,
            mock.sentinel.job_handle,
            mock.sentinel.job_info_class,
            self._ctypes.byref(mock.sentinel.job_info),
            self._ctypes.sizeof(mock.sentinel.job_info),
            kernel32_lib_func=True)

    def test_assign_process_to_job_object(self):
        self._procutils.assign_process_to_job_object(
            mock.sentinel.job_handle,
            mock.sentinel.process_handle)

        self._mock_run.assert_called_once_with(
            self._mock_kernel32.AssignProcessToJobObject,
            mock.sentinel.job_handle,
            mock.sentinel.process_handle,
            kernel32_lib_func=True)

    def test_open_process(self):
        process_handle = self._procutils.open_process(
            mock.sentinel.pid,
            mock.sentinel.desired_access,
            mock.sentinel.inherit_handle)

        self._mock_run.assert_called_once_with(
            self._mock_kernel32.OpenProcess,
            mock.sentinel.desired_access,
            mock.sentinel.inherit_handle,
            mock.sentinel.pid,
            error_ret_vals=[None],
            kernel32_lib_func=True)
        self.assertEqual(self._mock_run.return_value, process_handle)

    @ddt.data({},
              {'assign_job_exc': Exception})
    @ddt.unpack
    @mock.patch.object(processutils.ProcessUtils, 'open_process')
    @mock.patch.object(processutils.ProcessUtils, 'create_job_object')
    @mock.patch.object(processutils.ProcessUtils,
                       'set_information_job_object')
    @mock.patch.object(processutils.ProcessUtils,
                       'assign_process_to_job_object')
    @mock.patch.object(processutils.kernel32_struct,
                       'JOBOBJECT_EXTENDED_LIMIT_INFORMATION')
    def test_kill_process_on_job_close(self, mock_job_limit_struct,
                                       mock_assign_job,
                                       mock_set_job_info,
                                       mock_create_job,
                                       mock_open_process,
                                       assign_job_exc=None):
        mock_assign_job.side_effect = assign_job_exc
        mock_open_process.return_value = mock.sentinel.process_handle
        mock_create_job.return_value = mock.sentinel.job_handle

        if assign_job_exc:
            self.assertRaises(assign_job_exc,
                              self._procutils.kill_process_on_job_close,
                              mock.sentinel.pid)
        else:
            self._procutils.kill_process_on_job_close(mock.sentinel.pid)

        mock_open_process.assert_called_once_with(
            mock.sentinel.pid,
            w_const.PROCESS_SET_QUOTA | w_const.PROCESS_TERMINATE)
        mock_create_job.assert_called_once_with()

        mock_job_limit_struct.assert_called_once_with()
        mock_job_limit = mock_job_limit_struct.return_value
        self.assertEqual(w_const.JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE,
                         mock_job_limit.BasicLimitInformation.LimitFlags)

        mock_set_job_info.assert_called_once_with(
            mock.sentinel.job_handle,
            w_const.JobObjectExtendedLimitInformation,
            mock_job_limit)
        mock_assign_job.assert_called_once_with(
            mock.sentinel.job_handle,
            mock.sentinel.process_handle)

        exp_closed_handles = [mock.sentinel.process_handle]
        if assign_job_exc:
            exp_closed_handles.append(mock.sentinel.job_handle)

        self._win32_utils.close_handle.assert_has_calls(
            [mock.call(handle) for handle in exp_closed_handles])

    @ddt.data({},
              {'wait_exc': Exception})
    @ddt.unpack
    @mock.patch.object(processutils.ProcessUtils, 'open_process')
    def test_wait_for_multiple_processes(self, mock_open_process,
                                         wait_exc=None):
        pids = [mock.sentinel.pid0, mock.sentinel.pid1]
        phandles = [mock.sentinel.process_handle_0,
                    mock.sentinel.process_handle_1]

        mock_wait = self._win32_utils.wait_for_multiple_objects
        mock_wait.side_effect = wait_exc
        mock_open_process.side_effect = phandles

        if wait_exc:
            self.assertRaises(wait_exc,
                              self._procutils.wait_for_multiple_processes,
                              pids,
                              mock.sentinel.wait_all,
                              mock.sentinel.milliseconds)
        else:
            self._procutils.wait_for_multiple_processes(
                pids,
                mock.sentinel.wait_all,
                mock.sentinel.milliseconds)

        mock_open_process.assert_has_calls(
            [mock.call(pid,
                       desired_access=w_const.SYNCHRONIZE)
             for pid in pids])
        self._win32_utils.close_handle.assert_has_calls(
            [mock.call(handle) for handle in phandles])

        mock_wait.assert_called_once_with(phandles,
                                          mock.sentinel.wait_all,
                                          mock.sentinel.milliseconds)



# Copyright 2016 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base WMI utility class.
"""

import imp
import sys
import threading
import time

if sys.platform == 'win32':
    import wmi

from oslo_log import log as logging
from oslo_utils import reflection

LOG = logging.getLogger(__name__)


class BaseUtils(object):

    _WMI_CONS = {}

    def _get_wmi_obj(self, moniker, **kwargs):
        return wmi.WMI(moniker=moniker, **kwargs)

    def _get_wmi_conn(self, moniker, **kwargs):
        if sys.platform != 'win32':
            return None
        if kwargs:
            return self._get_wmi_obj(moniker, **kwargs)
        if moniker in self._WMI_CONS:
            return self._WMI_CONS[moniker]

        wmi_conn = self._get_wmi_obj(moniker)
        self._WMI_CONS[moniker] = wmi_conn
        return wmi_conn


class BaseUtilsVirt(BaseUtils):

    _wmi_namespace = '//%s/root/virtualization/v2'
    _os_version = None
    _old_wmi = None

    def __init__(self, host='.'):
        self._vs_man_svc_attr = None
        self._host = host
        self._conn_attr = None
        self._compat_conn_attr = None

    @property
    def _conn(self):
        if not self._conn_attr:
            self._conn_attr = self._get_wmi_conn(
                self._wmi_namespace % self._host)
        return self._conn_attr

    @property
    def _compat_conn(self):
        if not self._compat_conn_attr:
            if not BaseUtilsVirt._os_version:
                # hostutils cannot be used for this, it would end up in
                # a circular import.
                os_version = wmi.WMI().Win32_OperatingSystem()[0].Version
                BaseUtilsVirt._os_version = list(
                    map(int, os_version.split('.')))

            if BaseUtilsVirt._os_version >= [6, 3]:
                self._compat_conn_attr = self._conn
            else:
                self._compat_conn_attr = self._get_wmi_compat_conn(
                    moniker=self._wmi_namespace % self._host)

        return self._compat_conn_attr

    @property
    def _vs_man_svc(self):
        if self._vs_man_svc_attr:
            return self._vs_man_svc_attr

        vs_man_svc = self._compat_conn.Msvm_VirtualSystemManagementService()[0]
        if BaseUtilsVirt._os_version >= [6, 3]:
            # NOTE(claudiub): caching this property on Windows / Hyper-V Server
            # 2012 (using the old WMI) can lead to memory leaks. PyMI doesn't
            # have those issues, so we can safely cache it.
            self._vs_man_svc_attr = vs_man_svc
        return vs_man_svc

    def _get_wmi_compat_conn(self, moniker, **kwargs):
        # old WMI should be used on Windows / Hyper-V Server 2012 whenever
        # .GetText_ is used (e.g.: AddResourceSettings). PyMI's and WMI's
        # .GetText_ have different results.
        if not BaseUtilsVirt._old_wmi:
            old_wmi_path = "%s.py" % wmi.__path__[0]
            BaseUtilsVirt._old_wmi = imp.load_source('old_wmi', old_wmi_path)
        return BaseUtilsVirt._old_wmi.WMI(moniker=moniker, **kwargs)

    def _get_wmi_obj(self, moniker, compatibility_mode=False, **kwargs):
        if not BaseUtilsVirt._os_version:
            # hostutils cannot be used for this, it would end up in
            # a circular import.
            os_version = wmi.WMI().Win32_OperatingSystem()[0].Version
            BaseUtilsVirt._os_version = list(map(int, os_version.split('.')))

        if not compatibility_mode or BaseUtilsVirt._os_version >= [6, 3]:
            return wmi.WMI(moniker=moniker, **kwargs)
        return self._get_wmi_compat_conn(moniker=moniker, **kwargs)


class SynchronizedMeta(type):
    """Use an rlock to synchronize all class methods."""

    def __init__(cls, cls_name, bases, attrs):
        super(SynchronizedMeta, cls).__init__(cls_name, bases, attrs)
        rlock = threading.RLock()

        for attr_name in attrs:
            attr = getattr(cls, attr_name)
            if callable(attr):
                decorated = SynchronizedMeta._synchronize(
                    attr, cls_name, rlock)
                setattr(cls, attr_name, decorated)

    @staticmethod
    def _synchronize(func, cls_name, rlock):
        def wrapper(*args, **kwargs):
            f_qual_name = reflection.get_callable_name(func)

            t_request = time.time()
            try:
                with rlock:
                    t_acquire = time.time()
                    LOG.debug("Method %(method_name)s acquired rlock. "
                              "Waited %(time_wait)0.3fs",
                              dict(method_name=f_qual_name,
                                   time_wait=t_acquire - t_request))
                    return func(*args, **kwargs)
            finally:
                t_release = time.time()
                LOG.debug("Method %(method_name)s released rlock. "
                          "Held %(time_held)0.3fs",
                          dict(method_name=f_qual_name,
                               time_held=t_release - t_acquire))
        return wrapper



# Copyright 2013 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import ctypes
import os
import shutil
import sys
import tempfile

from oslo_log import log as logging
from oslo_utils import fileutils
import six

from os_win._i18n import _
from os_win import _utils
from os_win import exceptions
from os_win.utils import _acl_utils
from os_win.utils import win32utils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi import libs as w_lib
from os_win.utils.winapi.libs import advapi32 as advapi32_def
from os_win.utils.winapi import wintypes

kernel32 = w_lib.get_shared_lib_handle(w_lib.KERNEL32)

LOG = logging.getLogger(__name__)


class PathUtils(object):

    def __init__(self):
        self._win32_utils = win32utils.Win32Utils()
        self._acl_utils = _acl_utils.ACLUtils()

    def open(self, path, mode):
        """Wrapper on __builtin__.open used to simplify unit testing."""
        from six.moves import builtins
        return builtins.open(path, mode)

    def exists(self, path):
        return os.path.exists(path)

    def makedirs(self, path):
        os.makedirs(path)

    def remove(self, path):
        os.remove(path)

    def rename(self, src, dest):
        os.rename(src, dest)

    def copy_dir(self, src, dest):
        shutil.copytree(src, dest)

    def copyfile(self, src, dest):
        self.copy(src, dest)

    def copy(self, src, dest, fail_if_exists=True):
        """Copies a file to a specified location.

        :param fail_if_exists: if set to True, the method fails if the
                               destination path exists.
        """
        # With large files this is 2x-3x faster than shutil.copy(src, dest),
        # especially when copying to a UNC target.
        if os.path.isdir(dest):
            src_fname = os.path.basename(src)
            dest = os.path.join(dest, src_fname)

        try:
            self._win32_utils.run_and_check_output(
                kernel32.CopyFileW,
                ctypes.c_wchar_p(src),
                ctypes.c_wchar_p(dest),
                wintypes.BOOL(fail_if_exists),
                kernel32_lib_func=True)
        except exceptions.Win32Exception as exc:
            err_msg = _('The file copy from %(src)s to %(dest)s failed.'
                        'Exception: %(exc)s')
            raise IOError(err_msg % dict(src=src, dest=dest, exc=exc))

    def copy_folder_files(self, src_dir, dest_dir):
        """Copies the files of the given src_dir to dest_dir.

        It will ignore any nested folders.

        :param src_dir: Given folder from which to copy files.
        :param dest_dir: Folder to which to copy files.
        """

        self.check_create_dir(dest_dir)

        for fname in os.listdir(src_dir):
            src = os.path.join(src_dir, fname)
            # ignore subdirs.
            if os.path.isfile(src):
                self.copy(src, os.path.join(dest_dir, fname))

    def move_folder_files(self, src_dir, dest_dir):
        """Moves the files of the given src_dir to dest_dir.

        It will ignore any nested folders.

        :param src_dir: Given folder from which to move files.
        :param dest_dir: Folder to which to move files.
        """

        for fname in os.listdir(src_dir):
            src = os.path.join(src_dir, fname)
            # ignore subdirs.
            if os.path.isfile(src):
                self.rename(src, os.path.join(dest_dir, fname))

    @_utils.retry_decorator(exceptions=exceptions.OSWinException,
                            error_codes=[w_const.ERROR_DIR_IS_NOT_EMPTY])
    def rmtree(self, path):
        try:
            shutil.rmtree(path)
        except exceptions.WindowsError as ex:
            # NOTE(claudiub): convert it to an OSWinException in order to use
            # the retry_decorator.
            raise exceptions.OSWinException(six.text_type(ex),
                                            error_code=ex.winerror)

    def check_create_dir(self, path):
        if not self.exists(path):
            LOG.debug('Creating directory: %s', path)
            self.makedirs(path)

    def check_remove_dir(self, path):
        if self.exists(path):
            LOG.debug('Removing directory: %s', path)
            self.rmtree(path)

    def is_symlink(self, path):
        if sys.version_info >= (3, 2):
            return os.path.islink(path)

        file_attr = self._win32_utils.run_and_check_output(
            kernel32.GetFileAttributesW,
            path,
            error_ret_vals=[w_const.INVALID_FILE_ATTRIBUTES],
            kernel32_lib_func=True)

        return bool(os.path.isdir(path) and (
            file_attr & w_const.FILE_ATTRIBUTE_REPARSE_POINT))

    def create_sym_link(self, link, target, target_is_dir=True):
        """If target_is_dir is True, a junction will be created.

        NOTE: Junctions only work on same filesystem.
        """

        self._win32_utils.run_and_check_output(kernel32.CreateSymbolicLinkW,
                                               link,
                                               target,
                                               target_is_dir,
                                               kernel32_lib_func=True)

    def create_temporary_file(self, suffix=None, *args, **kwargs):
        fd, tmp_file_path = tempfile.mkstemp(suffix=suffix, *args, **kwargs)
        os.close(fd)
        return tmp_file_path

    @contextlib.contextmanager
    def temporary_file(self, suffix=None, *args, **kwargs):
        """Creates a random, temporary, closed file, returning the file's

        path. It's different from tempfile.NamedTemporaryFile which returns
        an open file descriptor.
        """

        tmp_file_path = None
        try:
            tmp_file_path = self.create_temporary_file(suffix, *args, **kwargs)
            yield tmp_file_path
        finally:
            if tmp_file_path:
                fileutils.delete_if_exists(tmp_file_path)

    def add_acl_rule(self, path, trustee_name,
                     access_rights, access_mode,
                     inheritance_flags=0):
        """Adds the requested access rule to a file or object.

        Can be used for granting/revoking access.
        """
        p_to_free = []

        try:
            sec_info = self._acl_utils.get_named_security_info(
                obj_name=path,
                obj_type=w_const.SE_FILE_OBJECT,
                security_info_flags=w_const.DACL_SECURITY_INFORMATION)
            p_to_free.append(sec_info['pp_sec_desc'].contents)

            access = advapi32_def.EXPLICIT_ACCESS()
            access.grfAccessPermissions = access_rights
            access.grfAccessMode = access_mode
            access.grfInheritance = inheritance_flags
            access.Trustee.TrusteeForm = w_const.TRUSTEE_IS_NAME
            access.Trustee.pstrName = ctypes.c_wchar_p(trustee_name)

            pp_new_dacl = self._acl_utils.set_entries_in_acl(
                entry_count=1,
                p_explicit_entry_list=ctypes.pointer(access),
                p_old_acl=sec_info['pp_dacl'].contents)
            p_to_free.append(pp_new_dacl.contents)

            self._acl_utils.set_named_security_info(
                obj_name=path,
                obj_type=w_const.SE_FILE_OBJECT,
                security_info_flags=w_const.DACL_SECURITY_INFORMATION,
                p_dacl=pp_new_dacl.contents)
        finally:
            for p in p_to_free:
                self._win32_utils.local_free(p)

    def copy_acls(self, source_path, dest_path):
        p_to_free = []

        try:
            sec_info_flags = w_const.DACL_SECURITY_INFORMATION
            sec_info = self._acl_utils.get_named_security_info(
                obj_name=source_path,
                obj_type=w_const.SE_FILE_OBJECT,
                security_info_flags=sec_info_flags)
            p_to_free.append(sec_info['pp_sec_desc'].contents)

            self._acl_utils.set_named_security_info(
                obj_name=dest_path,
                obj_type=w_const.SE_FILE_OBJECT,
                security_info_flags=sec_info_flags,
                p_dacl=sec_info['pp_dacl'].contents)
        finally:
            for p in p_to_free:
                self._win32_utils.local_free(p)



# Copyright 2015 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from eventlet import patcher
import pbr.version

from os_win.utils.winapi import libs as w_libs


__version__ = pbr.version.VersionInfo(
    'os_win').version_string()

if sys.platform == 'win32':
    import wmi
    # We need to make sure that WMI uses the unpatched threading module.
    wmi.threading = patcher.original('threading')

    # The following will set the argument and return value types for the
    # foreign functions used throughout os_win using ctypes.
    w_libs.register()



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import os

from eventlet import patcher
from oslo_log import log as logging

from os_win._i18n import _
from os_win import constants
from os_win import exceptions
from os_win.utils.io import ioutils
from os_win.utils.winapi import constants as w_const

threading = patcher.original('threading')
time = patcher.original('time')

LOG = logging.getLogger(__name__)


class NamedPipeHandler(object):
    """Handles asynchronous I/O operations on a specified named pipe."""

    _MAX_LOG_ROTATE_RETRIES = 5

    def __init__(self, pipe_name, input_queue=None, output_queue=None,
                 connect_event=None, log_file=None):
        self._pipe_name = pipe_name
        self._input_queue = input_queue
        self._output_queue = output_queue
        self._log_file_path = log_file

        self._connect_event = connect_event
        self._stopped = threading.Event()
        self._workers = []
        self._pipe_handle = None
        self._lock = threading.Lock()

        self._ioutils = ioutils.IOUtils()

        self._setup_io_structures()

    def start(self):
        try:
            self._open_pipe()

            if self._log_file_path:
                self._log_file_handle = open(self._log_file_path, 'ab', 1)

            jobs = [self._read_from_pipe]
            if self._input_queue and self._connect_event:
                jobs.append(self._write_to_pipe)

            for job in jobs:
                worker = threading.Thread(target=job)
                worker.setDaemon(True)
                worker.start()
                self._workers.append(worker)
        except Exception as err:
            msg = (_("Named pipe handler failed to initialize. "
                     "Pipe Name: %(pipe_name)s "
                     "Error: %(err)s") %
                   {'pipe_name': self._pipe_name,
                    'err': err})
            LOG.error(msg)
            self.stop()
            raise exceptions.OSWinException(msg)

    def stop(self):
        self._stopped.set()

        # If any worker has been spawned already, we rely on it to have
        # cleaned up the handles before ending its execution.
        # Note that we expect the caller to synchronize the start/stop calls.
        if not self._workers:
            self._cleanup_handles()

        for worker in self._workers:
            # It may happen that another IO request was issued right after
            # we've set the stopped event and canceled pending requests.
            # In this case, retrying will ensure that the IO workers are
            # stopped properly and that there are no more outstanding IO
            # operations.
            while (worker.is_alive() and
                   worker is not threading.current_thread()):
                self._cancel_io()
                worker.join(0.5)

        self._workers = []

    def _cleanup_handles(self):
        self._close_pipe()

        if self._log_file_handle:
            self._log_file_handle.close()
            self._log_file_handle = None

        if self._r_overlapped.hEvent:
            self._ioutils.close_handle(self._r_overlapped.hEvent)
            self._r_overlapped.hEvent = None

        if self._w_overlapped.hEvent:
            self._ioutils.close_handle(self._w_overlapped.hEvent)
            self._w_overlapped.hEvent = None

    def _setup_io_structures(self):
        self._r_buffer = self._ioutils.get_buffer(
            constants.SERIAL_CONSOLE_BUFFER_SIZE)
        self._w_buffer = self._ioutils.get_buffer(
            constants.SERIAL_CONSOLE_BUFFER_SIZE)

        self._r_overlapped = self._ioutils.get_new_overlapped_structure()
        self._w_overlapped = self._ioutils.get_new_overlapped_structure()

        self._r_completion_routine = self._ioutils.get_completion_routine(
            self._read_callback)
        self._w_completion_routine = self._ioutils.get_completion_routine()

        self._log_file_handle = None

    def _open_pipe(self):
        """Opens a named pipe in overlapped mode for asyncronous I/O."""
        self._ioutils.wait_named_pipe(self._pipe_name)

        self._pipe_handle = self._ioutils.open(
            self._pipe_name,
            desired_access=(w_const.GENERIC_READ | w_const.GENERIC_WRITE),
            share_mode=(w_const.FILE_SHARE_READ | w_const.FILE_SHARE_WRITE),
            creation_disposition=w_const.OPEN_EXISTING,
            flags_and_attributes=w_const.FILE_FLAG_OVERLAPPED)

    def _close_pipe(self):
        if self._pipe_handle:
            self._ioutils.close_handle(self._pipe_handle)
            self._pipe_handle = None

    def _cancel_io(self):
        if self._pipe_handle:
            # We ignore invalid handle errors. Even if the pipe is closed
            # and the handle reused, by specifying the overlapped structures
            # we ensure that we don't cancel IO operations other than the
            # ones that we care about.
            self._ioutils.cancel_io(self._pipe_handle, self._r_overlapped,
                                    ignore_invalid_handle=True)
            self._ioutils.cancel_io(self._pipe_handle, self._w_overlapped,
                                    ignore_invalid_handle=True)

    def _read_from_pipe(self):
        self._start_io_worker(self._ioutils.read,
                              self._r_buffer,
                              self._r_overlapped,
                              self._r_completion_routine)

    def _write_to_pipe(self):
        self._start_io_worker(self._ioutils.write,
                              self._w_buffer,
                              self._w_overlapped,
                              self._w_completion_routine,
                              self._get_data_to_write)

    def _start_io_worker(self, func, buff, overlapped_structure,
                         completion_routine, buff_update_func=None):
        try:
            while not self._stopped.isSet():
                if buff_update_func:
                    num_bytes = buff_update_func()
                    if not num_bytes:
                        continue
                else:
                    num_bytes = len(buff)

                func(self._pipe_handle, buff, num_bytes,
                     overlapped_structure, completion_routine)
        except Exception:
            self._stopped.set()
        finally:
            with self._lock:
                self._cleanup_handles()

    def _read_callback(self, num_bytes):
        data = self._ioutils.get_buffer_data(self._r_buffer,
                                             num_bytes)
        if self._output_queue:
            self._output_queue.put(data)

        if self._log_file_handle:
            self._write_to_log(data)

    def _get_data_to_write(self):
        while not (self._stopped.isSet() or self._connect_event.isSet()):
            time.sleep(1)

        data = self._input_queue.get()
        if data:
            self._ioutils.write_buffer_data(self._w_buffer, data)
            return len(data)
        return 0

    def _write_to_log(self, data):
        if self._stopped.isSet():
            return

        try:
            log_size = self._log_file_handle.tell() + len(data)
            if log_size >= constants.MAX_CONSOLE_LOG_FILE_SIZE:
                self._rotate_logs()
            self._log_file_handle.write(data)
        except Exception:
            self._stopped.set()

    def flush_log_file(self):
        try:
            self._log_file_handle.flush()
        except (AttributeError, ValueError):
            # We'll ignore errors caused by closed handles.
            pass

    def _rotate_logs(self):
        self._log_file_handle.flush()
        self._log_file_handle.close()

        log_archive_path = self._log_file_path + '.1'

        if os.path.exists(log_archive_path):
            self._retry_if_file_in_use(os.remove,
                                       log_archive_path)

        self._retry_if_file_in_use(os.rename,
                                   self._log_file_path,
                                   log_archive_path)

        self._log_file_handle = open(
            self._log_file_path, 'ab', 1)

    def _retry_if_file_in_use(self, f, *args, **kwargs):
        # The log files might be in use if the console log is requested
        # while a log rotation is attempted.
        retry_count = 0
        while True:
            try:
                return f(*args, **kwargs)
            except WindowsError as err:
                if (err.errno == errno.EACCES and
                        retry_count < self._MAX_LOG_ROTATE_RETRIES):
                    retry_count += 1
                    time.sleep(1)
                else:
                    raise



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Error codes and descriptions, as provided by iscsierr.h

from os_win._i18n import _
from os_win.utils.winapi import constants as w_const

err_msg_dict = {
    w_const.ISDSC_NON_SPECIFIC_ERROR:
        _('A non specific error occurred.'),
    w_const.ISDSC_LOGIN_FAILED:
        _('Login Failed.'),
    w_const.ISDSC_CONNECTION_FAILED:
        _('Connection Failed.'),
    w_const.ISDSC_INITIATOR_NODE_ALREADY_EXISTS:
        _('Initiator Node Already Exists.'),
    w_const.ISDSC_INITIATOR_NODE_NOT_FOUND:
        _('Initiator Node Does Not Exist.'),
    w_const.ISDSC_TARGET_MOVED_TEMPORARILY:
        _('Target Moved Temporarily.'),
    w_const.ISDSC_TARGET_MOVED_PERMANENTLY:
        _('Target Moved Permanently.'),
    w_const.ISDSC_INITIATOR_ERROR:
        _('Initiator Error.'),
    w_const.ISDSC_AUTHENTICATION_FAILURE:
        _('Authentication Failure.'),
    w_const.ISDSC_AUTHORIZATION_FAILURE:
        _('Authorization Failure.'),
    w_const.ISDSC_NOT_FOUND:
        _('Not Found.'),
    w_const.ISDSC_TARGET_REMOVED:
        _('Target Removed.'),
    w_const.ISDSC_UNSUPPORTED_VERSION:
        _('Unsupported Version.'),
    w_const.ISDSC_TOO_MANY_CONNECTIONS:
        _('Too many Connections.'),
    w_const.ISDSC_MISSING_PARAMETER:
        _('Missing Parameter.'),
    w_const.ISDSC_CANT_INCLUDE_IN_SESSION:
        _('Can not include in session.'),
    w_const.ISDSC_SESSION_TYPE_NOT_SUPPORTED:
        _('Session type not supported.'),
    w_const.ISDSC_TARGET_ERROR:
        _('Target Error.'),
    w_const.ISDSC_SERVICE_UNAVAILABLE:
        _('Service Unavailable.'),
    w_const.ISDSC_OUT_OF_RESOURCES:
        _('Out of Resources.'),
    w_const.ISDSC_CONNECTION_ALREADY_EXISTS:
        _('Connections already exist on initiator node.'),
    w_const.ISDSC_SESSION_ALREADY_EXISTS:
        _('Session Already Exists.'),
    w_const.ISDSC_INITIATOR_INSTANCE_NOT_FOUND:
        _('Initiator Instance Does Not Exist.'),
    w_const.ISDSC_TARGET_ALREADY_EXISTS:
        _('Target Already Exists.'),
    w_const.ISDSC_DRIVER_BUG:
        _('The iscsi driver implementation did '
          'not complete an operation correctly.'),
    w_const.ISDSC_INVALID_TEXT_KEY:
        _('An invalid key text was encountered.'),
    w_const.ISDSC_INVALID_SENDTARGETS_TEXT:
        _('Invalid SendTargets response text was encountered.'),
    w_const.ISDSC_INVALID_SESSION_ID:
        _('Invalid Session Id.'),
    w_const.ISDSC_SCSI_REQUEST_FAILED:
        _('The scsi request failed.'),
    w_const.ISDSC_TOO_MANY_SESSIONS:
        _('Exceeded max sessions for this initiator.'),
    w_const.ISDSC_SESSION_BUSY:
        _('Session is busy since a request is already in progress.'),
    w_const.ISDSC_TARGET_MAPPING_UNAVAILABLE:
        _('The target mapping requested is not available.'),
    w_const.ISDSC_ADDRESS_TYPE_NOT_SUPPORTED:
        _('The Target Address type given is not supported.'),
    w_const.ISDSC_LOGON_FAILED:
        _('Logon Failed.'),
    w_const.ISDSC_SEND_FAILED:
        _('TCP Send Failed.'),
    w_const.ISDSC_TRANSPORT_ERROR:
        _('TCP Transport Error'),
    w_const.ISDSC_VERSION_MISMATCH:
        _('iSCSI Version Mismatch'),
    w_const.ISDSC_TARGET_MAPPING_OUT_OF_RANGE:
        _('The Target Mapping Address passed is out of range for the '
          'adapter configuration.'),
    w_const.ISDSC_TARGET_PRESHAREDKEY_UNAVAILABLE:
        _('The preshared key for the target or IKE identification '
          'payload is not available.'),
    w_const.ISDSC_TARGET_AUTHINFO_UNAVAILABLE:
        _('The authentication information for '
          'the target is not available.'),
    w_const.ISDSC_TARGET_NOT_FOUND:
        _('The target name is not found or is '
          'marked as hidden from login.'),
    w_const.ISDSC_LOGIN_USER_INFO_BAD:
        _('One or more parameters specified in '
          'LoginTargetIN structure is invalid.'),
    w_const.ISDSC_TARGET_MAPPING_EXISTS:
        _('Given target mapping already exists.'),
    w_const.ISDSC_HBA_SECURITY_CACHE_FULL:
        _('The HBA security information cache '
          'is full.'),
    w_const.ISDSC_INVALID_PORT_NUMBER:
        _('The port number passed is '
          'not valid for the initiator.'),
    w_const.ISDSC_OPERATION_NOT_ALL_SUCCESS:
        _('The operation was not successful '
          'for all initiators or discovery '
          'methods.'),
    w_const.ISDSC_HBA_SECURITY_CACHE_NOT_SUPPORTED:
        _('The HBA security information '
          'cache is not supported by '
          'this adapter.'),
    w_const.ISDSC_IKE_ID_PAYLOAD_TYPE_NOT_SUPPORTED:
        _('The IKE id payload type '
          'specified is not supported.'),
    w_const.ISDSC_IKE_ID_PAYLOAD_INCORRECT_SIZE:
        _('The IKE id payload size '
          'specified is not correct.'),
    w_const.ISDSC_TARGET_PORTAL_ALREADY_EXISTS:
        _('Target Portal Structure '
          'Already Exists.'),
    w_const.ISDSC_TARGET_ADDRESS_ALREADY_EXISTS:
        _('Target Address Structure '
          'Already Exists.'),
    w_const.ISDSC_NO_AUTH_INFO_AVAILABLE:
        _('There is no IKE authentication '
          'information available.'),
    w_const.ISDSC_NO_TUNNEL_OUTER_MODE_ADDRESS:
        _('There is no tunnel mode outer '
          'address specified.'),
    w_const.ISDSC_CACHE_CORRUPTED:
        _('Authentication or tunnel '
          'address cache is corrupted.'),
    w_const.ISDSC_REQUEST_NOT_SUPPORTED:
        _('The request or operation '
          'is not supported.'),
    w_const.ISDSC_TARGET_OUT_OF_RESORCES:
        _('The target does not have enough '
          'resources to process the '
          'given request.'),
    w_const.ISDSC_SERVICE_DID_NOT_RESPOND:
        _('The initiator service did '
          'not respond to the request '
          'sent by the driver.'),
    w_const.ISDSC_ISNS_SERVER_NOT_FOUND:
        _('The Internet Storage Name Server (iSNS) '
          'server was not found or is unavailable.'),
    w_const.ISDSC_OPERATION_REQUIRES_REBOOT:
        _('The operation was successful but '
          'requires a driver reload or reboot '
          'to become effective.'),
    w_const.ISDSC_NO_PORTAL_SPECIFIED:
        _('There is no target portal available '
          'to complete the login.'),
    w_const.ISDSC_CANT_REMOVE_LAST_CONNECTION:
        _('Cannot remove the last '
          'connection for a session.'),
    w_const.ISDSC_SERVICE_NOT_RUNNING:
        _('The Microsoft iSCSI initiator '
          'service has not been started.'),
    w_const.ISDSC_TARGET_ALREADY_LOGGED_IN:
        _('The target has already been '
          'logged in via an iSCSI session.'),
    w_const.ISDSC_DEVICE_BUSY_ON_SESSION:
        _('The session cannot be logged out '
          'since a device on that session is '
          'currently being used.'),
    w_const.ISDSC_COULD_NOT_SAVE_PERSISTENT_LOGIN_DATA:
        _('Failed to save persistent '
          'login information.'),
    w_const.ISDSC_COULD_NOT_REMOVE_PERSISTENT_LOGIN_DATA:
        _('Failed to remove '
          'persistent login '
          'information.'),
    w_const.ISDSC_PORTAL_NOT_FOUND:
        _('The specified portal was not found.'),
    w_const.ISDSC_INITIATOR_NOT_FOUND:
        _('The specified initiator '
          'name was not found.'),
    w_const.ISDSC_DISCOVERY_MECHANISM_NOT_FOUND:
        _('The specified discovery '
          'mechanism was not found.'),
    w_const.ISDSC_IPSEC_NOT_SUPPORTED_ON_OS:
        _('iSCSI does not support IPSEC '
          'for this version of the OS.'),
    w_const.ISDSC_PERSISTENT_LOGIN_TIMEOUT:
        _('The iSCSI service timed out waiting '
          'for all persistent logins to '
          'complete.'),
    w_const.ISDSC_SHORT_CHAP_SECRET:
        _('The specified CHAP secret is less than '
          '96 bits and will not be usable for '
          'authenticating over non ipsec connections.'),
    w_const.ISDSC_EVALUATION_PEROID_EXPIRED:
        _('The evaluation period for the '
          'iSCSI initiator service has '
          'expired.'),
    w_const.ISDSC_INVALID_CHAP_SECRET:
        _('CHAP secret given does not conform '
          'to the standard. Please see system '
          'event log for more information.'),
    w_const.ISDSC_INVALID_TARGET_CHAP_SECRET:
        _('Target CHAP secret given is invalid. Maximum size of CHAP secret'
          'is 16 bytes. Minimum size is 12 bytes if IPSec is not used.'),
    w_const.ISDSC_INVALID_INITIATOR_CHAP_SECRET:
        _('Initiator CHAP secret given is '
          'invalid. Maximum size of CHAP '
          'secret is 16 bytes. Minimum size '
          'is 12 bytes if IPSec is '
          'not used.'),
    w_const.ISDSC_INVALID_CHAP_USER_NAME:
        _('CHAP Username given is invalid.'),
    w_const.ISDSC_INVALID_LOGON_AUTH_TYPE:
        _('Logon Authentication type '
          'given is invalid.'),
    w_const.ISDSC_INVALID_TARGET_MAPPING:
        _('Target Mapping information '
          'given is invalid.'),
    w_const.ISDSC_INVALID_TARGET_ID:
        _('Target Id given in '
          'Target Mapping is invalid.'),
    w_const.ISDSC_INVALID_ISCSI_NAME:
        _('The iSCSI name specified contains '
          'invalid characters or is too long.'),
    w_const.ISDSC_INCOMPATIBLE_ISNS_VERSION:
        _('The version number returned from the '
          'Internet Storage Name Server (iSNS) '
          'server is not compatible with this '
          'version of the iSNS client.'),
    w_const.ISDSC_FAILED_TO_CONFIGURE_IPSEC:
        _('Initiator failed to configure IPSec '
          'for the given connection. This could '
          'be because of low resources.'),
    w_const.ISDSC_BUFFER_TOO_SMALL:
        _('The buffer given for processing '
          'the request is too small.'),
    w_const.ISDSC_INVALID_LOAD_BALANCE_POLICY:
        _('The given Load Balance '
          'policy is not recognized '
          'by iScsi initiator.'),
    w_const.ISDSC_INVALID_PARAMETER:
        _('One or more paramaters '
          'specified is not valid.'),
    w_const.ISDSC_DUPLICATE_PATH_SPECIFIED:
        _('Duplicate PathIds were '
          'specified in the call to '
          'set Load Balance Policy.'),
    w_const.ISDSC_PATH_COUNT_MISMATCH:
        _('Number of paths specified in '
          'Set Load Balance Policy does not '
          'match the number of paths to the target.'),
    w_const.ISDSC_INVALID_PATH_ID:
        _('Path Id specified in the call to '
          'set Load Balance Policy is not valid'),
    w_const.ISDSC_MULTIPLE_PRIMARY_PATHS_SPECIFIED:
        _('Multiple primary paths '
          'specified when only one '
          'primary path is expected.'),
    w_const.ISDSC_NO_PRIMARY_PATH_SPECIFIED:
        _('No primary path specified when '
          'at least one is expected.'),
    w_const.ISDSC_DEVICE_ALREADY_PERSISTENTLY_BOUND:
        _('Device is already a '
          'persistently bound device.'),
    w_const.ISDSC_DEVICE_NOT_FOUND:
        _('Device was not found.'),
    w_const.ISDSC_DEVICE_NOT_ISCSI_OR_PERSISTENT:
        _('The device specified does not '
          'originate from an iSCSI disk '
          'or a persistent iSCSI login.'),
    w_const.ISDSC_DNS_NAME_UNRESOLVED:
        _('The DNS name specified was not resolved.'),
    w_const.ISDSC_NO_CONNECTION_AVAILABLE:
        _('There is no connection available '
          'in the iSCSI session to '
          'process the request.'),
    w_const.ISDSC_LB_POLICY_NOT_SUPPORTED:
        _('The given Load Balance '
          'policy is not supported.'),
    w_const.ISDSC_REMOVE_CONNECTION_IN_PROGRESS:
        _('A remove connection request '
          'is already in progress for '
          'this session.'),
    w_const.ISDSC_INVALID_CONNECTION_ID:
        _('Given connection was not '
          'found in the session.'),
    w_const.ISDSC_CANNOT_REMOVE_LEADING_CONNECTION:
        _('The leading connection in '
          'the session cannot be '
          'removed.'),
    w_const.ISDSC_RESTRICTED_BY_GROUP_POLICY:
        _('The operation cannot be performed '
          'since it does not conform with '
          'the group policy assigned to '
          'this computer.'),
    w_const.ISDSC_ISNS_FIREWALL_BLOCKED:
        _('The operation cannot be performed since '
          'the Internet Storage Name Server '
          '(iSNS) firewall exception has '
          'not been enabled.'),
    w_const.ISDSC_FAILURE_TO_PERSIST_LB_POLICY:
        _('Failed to persist load '
          'balancing policy parameters.'),
    w_const.ISDSC_INVALID_HOST:
        _('The name could not be resolved to an IP Address.'),
}



# Copyright 2014 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
from six.moves import range  # noqa

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils import _wqlutils
from os_win.utils.compute import vmutils


@ddt.ddt
class VMUtilsTestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the Hyper-V VMUtils class."""

    _autospec_classes = [
        vmutils.jobutils.JobUtils,
        vmutils.pathutils.PathUtils,
    ]

    _FAKE_VM_NAME = 'fake_vm'
    _FAKE_MEMORY_MB = 2
    _FAKE_VCPUS_NUM = 4
    _FAKE_JOB_PATH = 'fake_job_path'
    _FAKE_RET_VAL = 0
    _FAKE_PATH = "fake_path"
    _FAKE_CTRL_PATH = 'fake_ctrl_path'
    _FAKE_CTRL_ADDR = 0
    _FAKE_DRIVE_ADDR = 0
    _FAKE_MOUNTED_DISK_PATH = 'fake_mounted_disk_path'
    _FAKE_VM_PATH = "fake_vm_path"
    _FAKE_VHD_PATH = "fake_vhd_path"
    _FAKE_DVD_PATH = "fake_dvd_path"
    _FAKE_VOLUME_DRIVE_PATH = "fake_volume_drive_path"
    _FAKE_VM_UUID = "04e79212-39bc-4065-933c-50f6d48a57f6"
    _FAKE_INSTANCE = {"name": _FAKE_VM_NAME,
                      "uuid": _FAKE_VM_UUID}
    _FAKE_SNAPSHOT_PATH = "fake_snapshot_path"
    _FAKE_RES_DATA = "fake_res_data"
    _FAKE_HOST_RESOURCE = "fake_host_resource"
    _FAKE_CLASS = "FakeClass"
    _FAKE_RES_PATH = "fake_res_path"
    _FAKE_RES_NAME = 'fake_res_name'
    _FAKE_ADDRESS = "fake_address"
    _FAKE_DYNAMIC_MEMORY_RATIO = 1.0
    _FAKE_MONITOR_COUNT = 1

    _FAKE_MEMORY_INFO = {'DynamicMemoryEnabled': True,
                         'Reservation': 1024,
                         'Limit': 4096,
                         'Weight': 5000,
                         'MaxMemoryBlocksPerNumaNode': 2048}

    _FAKE_SUMMARY_INFO = {'NumberOfProcessors': 4,
                          'EnabledState': 2,
                          'MemoryUsage': 2,
                          'UpTime': 1}

    _DEFINE_SYSTEM = 'DefineSystem'
    _DESTROY_SYSTEM = 'DestroySystem'
    _DESTROY_SNAPSHOT = 'DestroySnapshot'
    _VM_GEN = constants.VM_GEN_2

    _VIRTUAL_SYSTEM_TYPE_REALIZED = 'Microsoft:Hyper-V:System:Realized'

    def setUp(self):
        super(VMUtilsTestCase, self).setUp()
        self._vmutils = vmutils.VMUtils()
        self._vmutils._conn_attr = mock.MagicMock()
        self._jobutils = self._vmutils._jobutils

    def test_get_vm_summary_info(self):
        self._lookup_vm()

        mock_summary = mock.MagicMock()
        mock_svc = self._vmutils._vs_man_svc
        mock_svc.GetSummaryInformation.return_value = (self._FAKE_RET_VAL,
                                                       [mock_summary])

        for (key, val) in self._FAKE_SUMMARY_INFO.items():
            setattr(mock_summary, key, val)

        summary = self._vmutils.get_vm_summary_info(self._FAKE_VM_NAME)
        self.assertEqual(self._FAKE_SUMMARY_INFO, summary)

    def _lookup_vm(self):
        mock_vm = mock.MagicMock()
        self._vmutils._lookup_vm_check = mock.MagicMock(
            return_value=mock_vm)
        mock_vm.path_.return_value = self._FAKE_VM_PATH
        return mock_vm

    def test_lookup_vm_ok(self):
        mock_vm = mock.MagicMock()
        self._vmutils._conn.Msvm_ComputerSystem.return_value = [mock_vm]
        vm = self._vmutils._lookup_vm_check(self._FAKE_VM_NAME, as_vssd=False)
        self.assertEqual(mock_vm, vm)

    def test_lookup_vm_multiple(self):
        mockvm = mock.MagicMock()
        self._vmutils._conn.Msvm_ComputerSystem.return_value = [mockvm, mockvm]
        self.assertRaises(exceptions.HyperVException,
                          self._vmutils._lookup_vm_check,
                          self._FAKE_VM_NAME,
                          as_vssd=False)

    def test_lookup_vm_none(self):
        self._vmutils._conn.Msvm_ComputerSystem.return_value = []
        self.assertRaises(exceptions.HyperVVMNotFoundException,
                          self._vmutils._lookup_vm_check,
                          self._FAKE_VM_NAME,
                          as_vssd=False)

    def test_lookup_vm_as_vssd(self):
        vssd = mock.MagicMock()
        expected_vssd = mock.MagicMock(
            VirtualSystemType=self._vmutils._VIRTUAL_SYSTEM_TYPE_REALIZED)

        self._vmutils._conn.Msvm_VirtualSystemSettingData.return_value = [
            vssd, expected_vssd]

        vssd = self._vmutils._lookup_vm_check(self._FAKE_VM_NAME)
        self.assertEqual(expected_vssd, vssd)

    @mock.patch.object(vmutils.VMUtils, '_lookup_vm')
    def test_vm_exists(self, mock_lookup_vm):
        result = self._vmutils.vm_exists(mock.sentinel.vm_name)

        self.assertTrue(result)
        mock_lookup_vm.assert_called_once_with(mock.sentinel.vm_name, False)

    def test_set_vm_memory_static(self):
        self._test_set_vm_memory_dynamic(dynamic_memory_ratio=1.0)

    def test_set_vm_memory_dynamic(self):
        self._test_set_vm_memory_dynamic(dynamic_memory_ratio=2.0)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_get_vm_memory_info(self, mock_get_element_associated_class):
        vmsetting = self._lookup_vm()

        mock_s = mock.MagicMock(**self._FAKE_MEMORY_INFO)
        mock_get_element_associated_class.return_value = [mock_s]

        memory = self._vmutils.get_vm_memory_info(self._FAKE_VM_NAME)
        self.assertEqual(self._FAKE_MEMORY_INFO, memory)

        mock_get_element_associated_class.assert_called_once_with(
            self._vmutils._compat_conn,
            self._vmutils._MEMORY_SETTING_DATA_CLASS,
            element_instance_id=vmsetting.InstanceID)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def _test_set_vm_memory_dynamic(self, mock_get_element_associated_class,
                                    dynamic_memory_ratio,
                                    mem_per_numa_node=None):
        mock_s = mock.MagicMock()

        mock_get_element_associated_class.return_value = [mock_s]

        self._vmutils._set_vm_memory(mock_s,
                                     self._FAKE_MEMORY_MB,
                                     mem_per_numa_node,
                                     dynamic_memory_ratio)

        self._vmutils._jobutils.modify_virt_resource.assert_called_once_with(
            mock_s)

        if mem_per_numa_node:
            self.assertEqual(mem_per_numa_node,
                             mock_s.MaxMemoryBlocksPerNumaNode)
        if dynamic_memory_ratio > 1:
            self.assertTrue(mock_s.DynamicMemoryEnabled)
        else:
            self.assertFalse(mock_s.DynamicMemoryEnabled)

    def test_set_vm_vcpus(self):
        self._check_set_vm_vcpus()

    def test_set_vm_vcpus_per_vnuma_node(self):
        self._check_set_vm_vcpus(vcpus_per_numa_node=1)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def _check_set_vm_vcpus(self, mock_get_element_associated_class,
                            vcpus_per_numa_node=None):
        procsetting = mock.MagicMock()
        mock_vmsettings = mock.MagicMock()
        mock_get_element_associated_class.return_value = [procsetting]

        self._vmutils._set_vm_vcpus(mock_vmsettings,
                                    self._FAKE_VCPUS_NUM,
                                    vcpus_per_numa_node,
                                    limit_cpu_features=False)

        self._vmutils._jobutils.modify_virt_resource.assert_called_once_with(
            procsetting)
        if vcpus_per_numa_node:
            self.assertEqual(vcpus_per_numa_node,
                             procsetting.MaxProcessorsPerNumaNode)
        mock_get_element_associated_class.assert_called_once_with(
            self._vmutils._conn, self._vmutils._PROCESSOR_SETTING_DATA_CLASS,
            element_instance_id=mock_vmsettings.InstanceID)

    def test_soft_shutdown_vm(self):
        mock_vm = self._lookup_vm()
        mock_shutdown = mock.MagicMock()
        mock_shutdown.InitiateShutdown.return_value = (self._FAKE_RET_VAL, )
        self._vmutils._conn.Msvm_ShutdownComponent.return_value = [
            mock_shutdown]

        self._vmutils.soft_shutdown_vm(self._FAKE_VM_NAME)

        mock_shutdown.InitiateShutdown.assert_called_once_with(
            Force=False, Reason=mock.ANY)
        self._vmutils._conn.Msvm_ShutdownComponent.assert_called_once_with(
            SystemName=mock_vm.Name)
        self._vmutils._jobutils.check_ret_val.assert_called_once_with(
            self._FAKE_RET_VAL, None)

    def test_soft_shutdown_vm_no_component(self):
        mock_vm = self._lookup_vm()
        self._vmutils._conn.Msvm_ShutdownComponent.return_value = []

        self._vmutils.soft_shutdown_vm(self._FAKE_VM_NAME)

        self._vmutils._conn.Msvm_ShutdownComponent.assert_called_once_with(
            SystemName=mock_vm.Name)
        self.assertFalse(self._vmutils._jobutils.check_ret_val.called)

    def test_get_vm_config_root_dir(self):
        mock_vm = self._lookup_vm()

        config_root_dir = self._vmutils.get_vm_config_root_dir(
            self._FAKE_VM_NAME)

        self.assertEqual(mock_vm.ConfigurationDataRoot, config_root_dir)

    @mock.patch.object(vmutils.VMUtils, '_get_vm_disks')
    @mock.patch.object(vmutils.VMUtils, '_lookup_vm_check')
    def test_get_vm_storage_paths(self, mock_lookup_vm_check,
                                  mock_get_vm_disks):
        mock_rasds = self._create_mock_disks()
        mock_get_vm_disks.return_value = ([mock_rasds[0]], [mock_rasds[1]])

        storage = self._vmutils.get_vm_storage_paths(self._FAKE_VM_NAME)
        (disk_files, volume_drives) = storage

        self.assertEqual([self._FAKE_VHD_PATH], disk_files)
        self.assertEqual([self._FAKE_VOLUME_DRIVE_PATH], volume_drives)
        mock_lookup_vm_check.assert_called_once_with(self._FAKE_VM_NAME)

    @mock.patch.object(vmutils.VMUtils, '_get_vm_disks')
    def test_get_vm_disks_by_instance_name(self, mock_get_vm_disks):
        self._lookup_vm()
        mock_get_vm_disks.return_value = mock.sentinel.vm_disks

        vm_disks = self._vmutils.get_vm_disks(self._FAKE_VM_NAME)

        self._vmutils._lookup_vm_check.assert_called_once_with(
            self._FAKE_VM_NAME)
        self.assertEqual(mock.sentinel.vm_disks, vm_disks)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_get_vm_disks(self, mock_get_element_associated_class):
        mock_vmsettings = self._lookup_vm()

        mock_rasds = self._create_mock_disks()
        mock_get_element_associated_class.return_value = mock_rasds

        (disks, volumes) = self._vmutils._get_vm_disks(mock_vmsettings)

        expected_calls = [
            mock.call(self._vmutils._conn,
                      self._vmutils._STORAGE_ALLOC_SETTING_DATA_CLASS,
                      element_instance_id=mock_vmsettings.InstanceID),
            mock.call(self._vmutils._conn,
                      self._vmutils._RESOURCE_ALLOC_SETTING_DATA_CLASS,
                      element_instance_id=mock_vmsettings.InstanceID)]

        mock_get_element_associated_class.assert_has_calls(expected_calls)

        self.assertEqual([mock_rasds[0]], disks)
        self.assertEqual([mock_rasds[1]], volumes)

    def _create_mock_disks(self):
        mock_rasd1 = mock.MagicMock()
        mock_rasd1.ResourceSubType = self._vmutils._HARD_DISK_RES_SUB_TYPE
        mock_rasd1.HostResource = [self._FAKE_VHD_PATH]
        mock_rasd1.Connection = [self._FAKE_VHD_PATH]
        mock_rasd1.Parent = self._FAKE_CTRL_PATH
        mock_rasd1.Address = self._FAKE_ADDRESS
        mock_rasd1.HostResource = [self._FAKE_VHD_PATH]

        mock_rasd2 = mock.MagicMock()
        mock_rasd2.ResourceSubType = self._vmutils._PHYS_DISK_RES_SUB_TYPE
        mock_rasd2.HostResource = [self._FAKE_VOLUME_DRIVE_PATH]

        return [mock_rasd1, mock_rasd2]

    def test_check_admin_permissions(self):
        mock_svc = self._vmutils._conn.Msvm_VirtualSystemManagementService
        mock_svc.return_value = False

        self.assertRaises(exceptions.HyperVAuthorizationException,
                          self._vmutils.check_admin_permissions)

    def test_set_nested_virtualization(self):
        self.assertRaises(NotImplementedError,
                          self._vmutils.set_nested_virtualization,
                          mock.sentinel.vm_name, mock.sentinel.state)

    @ddt.data(
        {'vnuma_enabled': mock.sentinel.vnuma_enabled},
        {'configuration_root_dir': mock.sentinel.configuration_root_dir},
        {'host_shutdown_action': mock.sentinel.shutdown_action},
        {})
    @ddt.unpack
    @mock.patch.object(vmutils.VMUtils, '_modify_virtual_system')
    @mock.patch.object(vmutils.VMUtils, '_set_vm_vcpus')
    @mock.patch.object(vmutils.VMUtils, '_set_vm_memory')
    @mock.patch.object(vmutils.VMUtils, '_set_vm_snapshot_type')
    @mock.patch.object(vmutils.VMUtils, '_lookup_vm_check')
    def test_update_vm(self, mock_lookup_vm_check,
                       mock_set_vm_snap_type,
                       mock_set_mem, mock_set_vcpus,
                       mock_modify_virtual_system,
                       host_shutdown_action=None,
                       configuration_root_dir=None, vnuma_enabled=None):
        mock_vmsettings = mock_lookup_vm_check.return_value
        self._vmutils.update_vm(
            mock.sentinel.vm_name, mock.sentinel.memory_mb,
            mock.sentinel.memory_per_numa, mock.sentinel.vcpus_num,
            mock.sentinel.vcpus_per_numa, mock.sentinel.limit_cpu_features,
            mock.sentinel.dynamic_mem_ratio, configuration_root_dir,
            host_shutdown_action=host_shutdown_action,
            vnuma_enabled=vnuma_enabled,
            snapshot_type=mock.sentinel.snap_type)

        mock_lookup_vm_check.assert_called_once_with(mock.sentinel.vm_name,
                                                     for_update=True)
        mock_set_mem.assert_called_once_with(
            mock_vmsettings, mock.sentinel.memory_mb,
            mock.sentinel.memory_per_numa, mock.sentinel.dynamic_mem_ratio)
        mock_set_vcpus.assert_called_once_with(
            mock_vmsettings, mock.sentinel.vcpus_num,
            mock.sentinel.vcpus_per_numa, mock.sentinel.limit_cpu_features)

        if configuration_root_dir:
            self.assertEqual(configuration_root_dir,
                             mock_vmsettings.ConfigurationDataRoot)
            self.assertEqual(configuration_root_dir,
                             mock_vmsettings.LogDataRoot)
            self.assertEqual(configuration_root_dir,
                             mock_vmsettings.SnapshotDataRoot)
            self.assertEqual(configuration_root_dir,
                             mock_vmsettings.SuspendDataRoot)
            self.assertEqual(configuration_root_dir,
                             mock_vmsettings.SwapFileDataRoot)
        if host_shutdown_action:
            self.assertEqual(host_shutdown_action,
                             mock_vmsettings.AutomaticShutdownAction)
        if vnuma_enabled:
            self.assertEqual(vnuma_enabled, mock_vmsettings.VirtualNumaEnabled)

        mock_set_vm_snap_type.assert_called_once_with(
            mock_vmsettings, mock.sentinel.snap_type)

        mock_modify_virtual_system.assert_called_once_with(
            mock_vmsettings)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_get_vm_scsi_controller(self, mock_get_element_associated_class):
        self._prepare_get_vm_controller(self._vmutils._SCSI_CTRL_RES_SUB_TYPE,
                                        mock_get_element_associated_class)
        path = self._vmutils.get_vm_scsi_controller(self._FAKE_VM_NAME)
        self.assertEqual(self._FAKE_RES_PATH, path)

    @mock.patch.object(vmutils.VMUtils, 'get_attached_disks')
    def test_get_free_controller_slot(self, mock_get_attached_disks):
        mock_disk = mock.MagicMock()
        mock_disk.AddressOnParent = 3
        mock_get_attached_disks.return_value = [mock_disk]

        response = self._vmutils.get_free_controller_slot(
            self._FAKE_CTRL_PATH)

        mock_get_attached_disks.assert_called_once_with(
            self._FAKE_CTRL_PATH)

        self.assertEqual(response, 0)

    def test_get_free_controller_slot_exception(self):
        fake_drive = mock.MagicMock()
        type(fake_drive).AddressOnParent = mock.PropertyMock(
            side_effect=list(range(constants.SCSI_CONTROLLER_SLOTS_NUMBER)))

        with mock.patch.object(
                self._vmutils,
                'get_attached_disks') as fake_get_attached_disks:
            fake_get_attached_disks.return_value = (
                [fake_drive] * constants.SCSI_CONTROLLER_SLOTS_NUMBER)
            self.assertRaises(exceptions.HyperVException,
                              self._vmutils.get_free_controller_slot,
                              mock.sentinel.scsi_controller_path)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_get_vm_ide_controller(self, mock_get_element_associated_class):
        self._prepare_get_vm_controller(
            self._vmutils._IDE_CTRL_RES_SUB_TYPE,
            mock_get_element_associated_class)
        path = self._vmutils.get_vm_ide_controller(
            mock.sentinel.FAKE_VM_SETTINGS, self._FAKE_ADDRESS)
        self.assertEqual(self._FAKE_RES_PATH, path)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_get_vm_ide_controller_none(self,
                                        mock_get_element_associated_class):
        self._prepare_get_vm_controller(
            self._vmutils._IDE_CTRL_RES_SUB_TYPE,
            mock_get_element_associated_class)
        path = self._vmutils.get_vm_ide_controller(
            mock.sentinel.FAKE_VM_SETTINGS, mock.sentinel.FAKE_NOT_FOUND_ADDR)
        self.assertNotEqual(self._FAKE_RES_PATH, path)

    def _prepare_get_vm_controller(self, resource_sub_type,
                                   mock_get_element_associated_class):
        self._lookup_vm()
        mock_rasds = mock.MagicMock()
        mock_rasds.path_.return_value = self._FAKE_RES_PATH
        mock_rasds.ResourceSubType = resource_sub_type
        mock_rasds.Address = self._FAKE_ADDRESS
        mock_get_element_associated_class.return_value = [mock_rasds]

    @mock.patch.object(vmutils.VMUtils, '_get_wmi_obj')
    def test_get_ide_ctrl_addr(self, mock_get_wmi_obj):
        mock_rasds = mock.Mock()
        mock_rasds.ResourceSubType = self._vmutils._IDE_CTRL_RES_SUB_TYPE
        mock_rasds.Address = mock.sentinel.ctrl_addr
        mock_get_wmi_obj.return_value = mock_rasds

        ret_val = self._vmutils._get_disk_ctrl_addr(mock.sentinel.ctrl_path)
        self.assertEqual(mock.sentinel.ctrl_addr, ret_val)

        mock_get_wmi_obj.assert_called_once_with(mock.sentinel.ctrl_path)

    @mock.patch.object(vmutils.VMUtils, '_get_vm_disk_controllers')
    @mock.patch.object(vmutils.VMUtils, '_get_wmi_obj')
    def test_get_scsi_ctrl_addr(self, mock_get_wmi_obj, mock_get_ctrls):
        mock_rasds = mock.Mock()
        mock_rasds.ResourceSubType = self._vmutils._SCSI_CTRL_RES_SUB_TYPE
        mock_rasds.associators.return_value = [mock.sentinel.vmsettings]
        mock_get_wmi_obj.return_value = mock_rasds

        mock_scsi_ctrls = ['someCtrl', self._FAKE_CTRL_PATH.upper(),
                           'someOtherCtrl']
        exp_ctrl_addr = 1

        mock_scsi_ctrl = mock.Mock()
        mock_scsi_ctrl.path_.side_effect = mock_scsi_ctrls
        mock_get_ctrls.return_value = [mock_scsi_ctrl] * len(mock_scsi_ctrls)

        ret_val = self._vmutils._get_disk_ctrl_addr(self._FAKE_CTRL_PATH)
        self.assertEqual(exp_ctrl_addr, ret_val)

        mock_get_wmi_obj.assert_called_once_with(self._FAKE_CTRL_PATH)
        mock_get_ctrls.assert_called_once_with(
            mock.sentinel.vmsettings, self._vmutils._SCSI_CTRL_RES_SUB_TYPE)

    @mock.patch.object(vmutils.VMUtils, 'get_free_controller_slot')
    @mock.patch.object(vmutils.VMUtils, '_get_vm_scsi_controller')
    def test_attach_scsi_drive(self, mock_get_vm_scsi_controller,
                               mock_get_free_controller_slot):
        mock_vm = self._lookup_vm()
        mock_get_vm_scsi_controller.return_value = self._FAKE_CTRL_PATH
        mock_get_free_controller_slot.return_value = self._FAKE_DRIVE_ADDR

        with mock.patch.object(self._vmutils,
                               'attach_drive') as mock_attach_drive:
            self._vmutils.attach_scsi_drive(mock_vm, self._FAKE_PATH,
                                            constants.DISK)

            mock_get_vm_scsi_controller.assert_called_once_with(mock_vm)
            mock_get_free_controller_slot.assert_called_once_with(
                self._FAKE_CTRL_PATH)
            mock_attach_drive.assert_called_once_with(
                mock_vm, self._FAKE_PATH, self._FAKE_CTRL_PATH,
                self._FAKE_DRIVE_ADDR, constants.DISK)

    @mock.patch.object(vmutils.VMUtils, 'attach_drive')
    @mock.patch.object(vmutils.VMUtils, '_get_vm_ide_controller')
    def test_attach_ide_drive(self, mock_get_ide_ctrl, mock_attach_drive):
        mock_vm = self._lookup_vm()

        self._vmutils.attach_ide_drive(self._FAKE_VM_NAME,
                                       self._FAKE_CTRL_PATH,
                                       self._FAKE_CTRL_ADDR,
                                       self._FAKE_DRIVE_ADDR)

        mock_get_ide_ctrl.assert_called_with(mock_vm, self._FAKE_CTRL_ADDR)
        mock_attach_drive.assert_called_once_with(
            self._FAKE_VM_NAME, self._FAKE_CTRL_PATH,
            mock_get_ide_ctrl.return_value, self._FAKE_DRIVE_ADDR,
            constants.DISK)

    @ddt.data(constants.DISK, constants.DVD)
    @mock.patch.object(vmutils.VMUtils, '_get_new_resource_setting_data')
    def test_attach_drive(self, drive_type, mock_get_new_rsd):
        mock_vm = self._lookup_vm()

        mock_drive_res = mock.Mock()
        mock_disk_res = mock.Mock()

        mock_get_new_rsd.side_effect = [mock_drive_res, mock_disk_res]
        self._jobutils.add_virt_resource.side_effect = [
            [mock.sentinel.drive_res_path],
            [mock.sentinel.disk_res_path]]

        self._vmutils.attach_drive(mock.sentinel.vm_name,
                                   mock.sentinel.disk_path,
                                   mock.sentinel.ctrl_path,
                                   mock.sentinel.drive_addr,
                                   drive_type)

        self._vmutils._lookup_vm_check.assert_called_once_with(
            mock.sentinel.vm_name, as_vssd=False)

        if drive_type == constants.DISK:
            exp_res_sub_types = [self._vmutils._DISK_DRIVE_RES_SUB_TYPE,
                                 self._vmutils._HARD_DISK_RES_SUB_TYPE]
        else:
            exp_res_sub_types = [self._vmutils._DVD_DRIVE_RES_SUB_TYPE,
                                 self._vmutils._DVD_DISK_RES_SUB_TYPE]

        mock_get_new_rsd.assert_has_calls(
            [mock.call(exp_res_sub_types[0]),
             mock.call(exp_res_sub_types[1],
                       self._vmutils._STORAGE_ALLOC_SETTING_DATA_CLASS)])

        self.assertEqual(mock.sentinel.ctrl_path, mock_drive_res.Parent)
        self.assertEqual(mock.sentinel.drive_addr, mock_drive_res.Address)
        self.assertEqual(mock.sentinel.drive_addr,
                         mock_drive_res.AddressOnParent)

        self.assertEqual(mock.sentinel.drive_res_path,
                         mock_disk_res.Parent)
        self.assertEqual([mock.sentinel.disk_path],
                         mock_disk_res.HostResource)

        self._jobutils.add_virt_resource.assert_has_calls(
            [mock.call(mock_drive_res, mock_vm),
             mock.call(mock_disk_res, mock_vm)])

    @mock.patch.object(vmutils.VMUtils, '_get_wmi_obj')
    @mock.patch.object(vmutils.VMUtils, '_get_new_resource_setting_data')
    def test_attach_drive_exc(self, mock_get_new_rsd, mock_get_wmi_obj):
        self._lookup_vm()

        mock_drive_res = mock.Mock()
        mock_disk_res = mock.Mock()

        mock_get_new_rsd.side_effect = [mock_drive_res, mock_disk_res]
        self._jobutils.add_virt_resource.side_effect = [
            [mock.sentinel.drive_res_path],
            exceptions.OSWinException]
        mock_get_wmi_obj.return_value = mock.sentinel.attached_drive_res

        self.assertRaises(exceptions.OSWinException,
                          self._vmutils.attach_drive,
                          mock.sentinel.vm_name,
                          mock.sentinel.disk_path,
                          mock.sentinel.ctrl_path,
                          mock.sentinel.drive_addr,
                          constants.DISK)

        mock_get_wmi_obj.assert_called_once_with(mock.sentinel.drive_res_path)
        self._jobutils.remove_virt_resource.assert_called_once_with(
            mock.sentinel.attached_drive_res)

    @mock.patch.object(vmutils.VMUtils,
                       '_get_mounted_disk_resource_from_path')
    def test_get_disk_attachment_info_detached(self, mock_get_disk_res):
        mock_get_disk_res.return_value = None
        self.assertRaises(exceptions.DiskNotFound,
                          self._vmutils.get_disk_attachment_info,
                          mock.sentinel.disk_path,
                          mock.sentinel.is_physical,
                          mock.sentinel.serial)

        mock_get_disk_res.assert_called_once_with(
            mock.sentinel.disk_path,
            mock.sentinel.is_physical,
            serial=mock.sentinel.serial)

    @ddt.data(True, False)
    @mock.patch.object(vmutils.VMUtils,
                       '_get_mounted_disk_resource_from_path')
    @mock.patch.object(vmutils.VMUtils,
                       '_get_disk_controller_type')
    @mock.patch.object(vmutils.VMUtils,
                       '_get_wmi_obj')
    @mock.patch.object(vmutils.VMUtils, '_get_disk_ctrl_addr')
    def test_get_disk_attachment_info(self, is_physical,
                                      mock_get_disk_ctrl_addr,
                                      mock_get_wmi_obj,
                                      mock_get_disk_ctrl_type,
                                      mock_get_disk_res):
        mock_res = mock_get_disk_res.return_value
        exp_res = mock_res if is_physical else mock_get_wmi_obj.return_value

        fake_slot = 5
        exp_res.AddressOnParent = str(fake_slot)

        exp_att_info = dict(
            controller_slot=fake_slot,
            controller_path=exp_res.Parent,
            controller_type=mock_get_disk_ctrl_type.return_value,
            controller_addr=mock_get_disk_ctrl_addr.return_value)

        att_info = self._vmutils.get_disk_attachment_info(
            mock.sentinel.disk_path,
            is_physical)
        self.assertEqual(exp_att_info, att_info)

        if not is_physical:
            mock_get_wmi_obj.assert_called_once_with(mock_res.Parent)
        mock_get_disk_ctrl_type.assert_called_once_with(exp_res.Parent)
        mock_get_disk_ctrl_addr.assert_called_once_with(exp_res.Parent)

    @ddt.data(vmutils.VMUtils._SCSI_CTRL_RES_SUB_TYPE,
              vmutils.VMUtils._IDE_CTRL_RES_SUB_TYPE)
    @mock.patch.object(vmutils.VMUtils, '_get_wmi_obj')
    def test_get_disk_controller_type(self, res_sub_type, mock_get_wmi_obj):
        mock_ctrl = mock_get_wmi_obj.return_value
        mock_ctrl.ResourceSubType = res_sub_type

        exp_ctrl_type = self._vmutils._disk_ctrl_type_mapping[res_sub_type]

        ctrl_type = self._vmutils._get_disk_controller_type(
            mock.sentinel.ctrl_path)
        self.assertEqual(exp_ctrl_type, ctrl_type)

        mock_get_wmi_obj.assert_called_once_with(mock.sentinel.ctrl_path)

    @mock.patch.object(vmutils.VMUtils, '_get_new_resource_setting_data')
    def test_create_scsi_controller(self, mock_get_new_rsd):
        mock_vm = self._lookup_vm()

        self._vmutils.create_scsi_controller(self._FAKE_VM_NAME)

        self._vmutils._jobutils.add_virt_resource.assert_called_once_with(
            mock_get_new_rsd.return_value, mock_vm)

    @mock.patch.object(vmutils.VMUtils, '_get_new_resource_setting_data')
    @mock.patch.object(vmutils.VMUtils, '_get_wmi_obj')
    def _test_attach_volume_to_controller(self, mock_get_wmi_obj,
                                          mock_get_new_rsd, disk_serial=None):
        mock_vm = self._lookup_vm()
        mock_diskdrive = mock.MagicMock()
        jobutils = self._vmutils._jobutils
        jobutils.add_virt_resource.return_value = [mock_diskdrive]
        mock_get_wmi_obj.return_value = mock_diskdrive

        self._vmutils.attach_volume_to_controller(
            self._FAKE_VM_NAME, self._FAKE_CTRL_PATH, self._FAKE_CTRL_ADDR,
            self._FAKE_MOUNTED_DISK_PATH, serial=disk_serial)

        self._vmutils._jobutils.add_virt_resource.assert_called_once_with(
            mock_get_new_rsd.return_value, mock_vm)

        if disk_serial:
            jobutils.modify_virt_resource.assert_called_once_with(
                mock_diskdrive)
            self.assertEqual(disk_serial, mock_diskdrive.ElementName)

    def test_attach_volume_to_controller_without_disk_serial(self):
        self._test_attach_volume_to_controller()

    def test_attach_volume_to_controller_with_disk_serial(self):
        self._test_attach_volume_to_controller(
            disk_serial=mock.sentinel.serial)

    @mock.patch.object(vmutils.VMUtils, '_get_new_setting_data')
    def test_create_nic(self, mock_get_new_virt_res):
        mock_vm = self._lookup_vm()
        mock_nic = mock_get_new_virt_res.return_value

        self._vmutils.create_nic(
            self._FAKE_VM_NAME, self._FAKE_RES_NAME, self._FAKE_ADDRESS)

        self._vmutils._jobutils.add_virt_resource.assert_called_once_with(
            mock_nic, mock_vm)

    def test_get_nic_data_by_name(self):
        nic_cls = self._vmutils._conn.Msvm_SyntheticEthernetPortSettingData
        nic_cls.return_value = [mock.sentinel.nic]

        nic = self._vmutils._get_nic_data_by_name(mock.sentinel.name)

        self.assertEqual(mock.sentinel.nic, nic)
        nic_cls.assert_called_once_with(ElementName=mock.sentinel.name)

    def test_get_missing_nic_data_by_name(self):
        nic_cls = self._vmutils._conn.Msvm_SyntheticEthernetPortSettingData
        nic_cls.return_value = []
        self.assertRaises(
            exceptions.HyperVvNicNotFound,
            self._vmutils._get_nic_data_by_name,
            mock.sentinel.name)

    @mock.patch.object(vmutils.VMUtils, '_get_nic_data_by_name')
    def test_destroy_nic(self, mock_get_nic_data_by_name):
        mock_nic_data = mock_get_nic_data_by_name.return_value

        # We expect this exception to be ignored.
        self._vmutils._jobutils.remove_virt_resource.side_effect = (
            exceptions.NotFound(message='fake_exc'))

        self._vmutils.destroy_nic(self._FAKE_VM_NAME,
                                  mock.sentinel.FAKE_NIC_NAME)

        self._vmutils._jobutils.remove_virt_resource.assert_called_once_with(
            mock_nic_data)

    @mock.patch.object(vmutils.VMUtils, '_lookup_vm_check')
    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_get_vm_nics(self, mock_get_assoc, mock_lookup_vm):
        vnics = self._vmutils._get_vm_nics(mock.sentinel.vm_name)

        self.assertEqual(mock_get_assoc.return_value, vnics)
        mock_lookup_vm.assert_called_once_with(mock.sentinel.vm_name)
        mock_get_assoc.assert_called_once_with(
            self._vmutils._compat_conn,
            self._vmutils._SYNTHETIC_ETHERNET_PORT_SETTING_DATA_CLASS,
            element_instance_id=mock_lookup_vm.return_value.InstanceId)

    @mock.patch.object(vmutils.VMUtils, '_get_vm_nics')
    def test_get_vm_nic_names(self, mock_get_vm_nics):
        exp_nic_names = ['port1', 'port2']

        mock_get_vm_nics.return_value = [
            mock.Mock(ElementName=nic_name)
            for nic_name in exp_nic_names]
        nic_names = self._vmutils.get_vm_nic_names(mock.sentinel.vm_name)

        self.assertEqual(exp_nic_names, nic_names)
        mock_get_vm_nics.assert_called_once_with(mock.sentinel.vm_name)

    def test_set_vm_state(self):
        mock_vm = self._lookup_vm()
        mock_vm.RequestStateChange.return_value = (
            self._FAKE_JOB_PATH, self._FAKE_RET_VAL)

        self._vmutils.set_vm_state(self._FAKE_VM_NAME,
                                   constants.HYPERV_VM_STATE_ENABLED)
        mock_vm.RequestStateChange.assert_called_with(
            constants.HYPERV_VM_STATE_ENABLED)

    def test_destroy_vm(self):
        self._lookup_vm()

        mock_svc = self._vmutils._vs_man_svc
        getattr(mock_svc, self._DESTROY_SYSTEM).return_value = (
            self._FAKE_JOB_PATH, self._FAKE_RET_VAL)

        self._vmutils.destroy_vm(self._FAKE_VM_NAME)

        getattr(mock_svc, self._DESTROY_SYSTEM).assert_called_with(
            self._FAKE_VM_PATH)

    @mock.patch.object(vmutils.VMUtils, 'get_vm_disks')
    def test_get_vm_physical_disk_mapping(self, mock_get_vm_disks):
        mock_phys_disk = self._create_mock_disks()[1]

        expected_serial = mock_phys_disk.ElementName
        expected_mapping = {
            expected_serial: {
                'resource_path': mock_phys_disk.path_.return_value,
                'mounted_disk_path': mock_phys_disk.HostResource[0]
            }
        }

        mock_get_vm_disks.return_value = ([], [mock_phys_disk])

        result = self._vmutils.get_vm_physical_disk_mapping(self._FAKE_VM_NAME)
        self.assertEqual(expected_mapping, result)
        mock_get_vm_disks.assert_called_once_with(self._FAKE_VM_NAME)

    @mock.patch.object(vmutils.VMUtils, '_get_wmi_obj')
    def test_set_disk_host_res(self, mock_get_wmi_obj):
        mock_diskdrive = mock_get_wmi_obj.return_value

        self._vmutils.set_disk_host_res(self._FAKE_RES_PATH,
                                        self._FAKE_MOUNTED_DISK_PATH)

        self._vmutils._jobutils.modify_virt_resource.assert_called_once_with(
            mock_diskdrive)

        mock_get_wmi_obj.assert_called_once_with(self._FAKE_RES_PATH, True)
        self.assertEqual(mock_diskdrive.HostResource,
                         [self._FAKE_MOUNTED_DISK_PATH])

    @mock.patch.object(vmutils.VMUtils, '_modify_virtual_system')
    @ddt.data(None, mock.sentinel.snap_name)
    def test_take_vm_snapshot(self, snap_name, mock_modify_virtual_system):
        self._lookup_vm()
        mock_snap = mock.Mock(ElementName=mock.sentinel.default_snap_name)

        mock_svc = self._get_snapshot_service()
        mock_svc.CreateSnapshot.return_value = (self._FAKE_JOB_PATH,
                                                mock.MagicMock(),
                                                self._FAKE_RET_VAL)

        mock_job = self._vmutils._jobutils.check_ret_val.return_value
        mock_job.associators.return_value = [mock_snap]

        snap_path = self._vmutils.take_vm_snapshot(self._FAKE_VM_NAME,
                                                   snap_name)

        self.assertEqual(mock_snap.path_.return_value, snap_path)
        mock_svc.CreateSnapshot.assert_called_with(
            AffectedSystem=self._FAKE_VM_PATH,
            SnapshotType=self._vmutils._SNAPSHOT_FULL)

        self._vmutils._jobutils.check_ret_val.assert_called_once_with(
            self._FAKE_RET_VAL, self._FAKE_JOB_PATH)

        mock_job.associators.assert_called_once_with(
            wmi_result_class=self._vmutils._VIRTUAL_SYSTEM_SETTING_DATA_CLASS,
            wmi_association_class=self._vmutils._AFFECTED_JOB_ELEMENT_CLASS)

        if snap_name:
            self.assertEqual(snap_name, mock_snap.ElementName)
            mock_modify_virtual_system.assert_called_once_with(mock_snap)
        else:
            self.assertEqual(mock.sentinel.default_snap_name,
                             mock_snap.ElementName)
            mock_modify_virtual_system.assert_not_called()

    @ddt.data(None, mock.sentinel.snap1)
    def test_get_vm_snapshots(self, snap_name):
        mock_snap1 = mock.Mock(ElementName=mock.sentinel.snap1)
        mock_snap2 = mock.Mock(ElementName=mock.sentinel.snap2)

        mock_vm = self._lookup_vm()
        mock_vm.associators.return_value = [mock_snap1, mock_snap2]

        snaps = self._vmutils.get_vm_snapshots(mock.sentinel.vm_name,
                                               snap_name)

        expected_snaps = [mock_snap1.path_.return_value]
        if not snap_name:
            expected_snaps += [mock_snap2.path_.return_value]

        self.assertEqual(expected_snaps, snaps)

        mock_vm.associators.assert_called_once_with(
            wmi_association_class=(
                self._vmutils._VIRTUAL_SYSTEM_SNAP_ASSOC_CLASS),
            wmi_result_class=(
                self._vmutils._VIRTUAL_SYSTEM_SETTING_DATA_CLASS))

    def test_remove_vm_snapshot(self):
        mock_svc = self._get_snapshot_service()
        getattr(mock_svc, self._DESTROY_SNAPSHOT).return_value = (
            self._FAKE_JOB_PATH, self._FAKE_RET_VAL)

        self._vmutils.remove_vm_snapshot(self._FAKE_SNAPSHOT_PATH)
        getattr(mock_svc, self._DESTROY_SNAPSHOT).assert_called_with(
            self._FAKE_SNAPSHOT_PATH)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_get_vm_dvd_disk_paths(self, mock_get_element_associated_class):
        self._lookup_vm()
        mock_sasd1 = mock.MagicMock(
            ResourceSubType=self._vmutils._DVD_DISK_RES_SUB_TYPE,
            HostResource=[mock.sentinel.FAKE_DVD_PATH1])
        mock_get_element_associated_class.return_value = [mock_sasd1]

        ret_val = self._vmutils.get_vm_dvd_disk_paths(self._FAKE_VM_NAME)
        self.assertEqual(mock.sentinel.FAKE_DVD_PATH1, ret_val[0])

    @mock.patch.object(vmutils.VMUtils,
                       '_get_mounted_disk_resource_from_path')
    def test_is_disk_attached(self, mock_get_mounted_disk_from_path):
        is_physical = True

        is_attached = self._vmutils.is_disk_attached(mock.sentinel.disk_path,
                                                     is_physical=is_physical)

        self.assertTrue(is_attached)
        mock_get_mounted_disk_from_path.assert_called_once_with(
            mock.sentinel.disk_path, is_physical)

    def test_detach_vm_disk(self):
        mock_disk = self._prepare_mock_disk()

        self._vmutils.detach_vm_disk(self._FAKE_VM_NAME,
                                     self._FAKE_HOST_RESOURCE,
                                     serial=mock.sentinel.serial)
        self._vmutils._jobutils.remove_virt_resource.assert_called_once_with(
            mock_disk)

    @ddt.data(None, mock.sentinel.serial)
    def test_get_mounted_disk_resource_from_path(self, serial):
        mock_disk = mock.MagicMock()

        if serial:
            self._vmutils._conn.query.return_value = [mock_disk]
        else:
            mock_disk.HostResource = [self._FAKE_MOUNTED_DISK_PATH]
            self._vmutils._conn.query.return_value = [
                mock.MagicMock(), mock_disk]

        physical_disk = self._vmutils._get_mounted_disk_resource_from_path(
            self._FAKE_MOUNTED_DISK_PATH, True, serial=serial)

        self.assertEqual(mock_disk, physical_disk)

    def test_get_controller_volume_paths(self):
        self._prepare_mock_disk()
        mock_disks = {self._FAKE_RES_PATH: self._FAKE_HOST_RESOURCE}
        disks = self._vmutils.get_controller_volume_paths(self._FAKE_RES_PATH)
        self.assertEqual(mock_disks, disks)

    def _prepare_mock_disk(self):
        mock_disk = mock.MagicMock()
        mock_disk.HostResource = [self._FAKE_HOST_RESOURCE]
        mock_disk.path.return_value.RelPath = self._FAKE_RES_PATH
        mock_disk.ResourceSubType = self._vmutils._HARD_DISK_RES_SUB_TYPE
        self._vmutils._conn.query.return_value = [mock_disk]

        return mock_disk

    def _get_snapshot_service(self):
        return self._vmutils._conn.Msvm_VirtualSystemSnapshotService()[0]

    def test_get_active_instances(self):
        fake_vm = mock.MagicMock()

        type(fake_vm).ElementName = mock.PropertyMock(
            side_effect=['active_vm', 'inactive_vm'])
        type(fake_vm).EnabledState = mock.PropertyMock(
            side_effect=[constants.HYPERV_VM_STATE_ENABLED,
                         constants.HYPERV_VM_STATE_DISABLED])
        self._vmutils.list_instances = mock.MagicMock(
            return_value=[mock.sentinel.fake_vm_name] * 2)
        self._vmutils._lookup_vm = mock.MagicMock(side_effect=[fake_vm] * 2)
        active_instances = self._vmutils.get_active_instances()

        self.assertEqual(['active_vm'], active_instances)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_get_vm_serial_ports(self, mock_get_element_associated_class):
        mock_vmsettings = self._lookup_vm()

        fake_serial_port = mock.MagicMock()
        fake_serial_port.ResourceSubType = (
            self._vmutils._SERIAL_PORT_RES_SUB_TYPE)

        mock_rasds = [fake_serial_port]
        mock_get_element_associated_class.return_value = mock_rasds

        ret_val = self._vmutils._get_vm_serial_ports(mock_vmsettings)

        self.assertEqual(mock_rasds, ret_val)
        mock_get_element_associated_class.assert_called_once_with(
            self._vmutils._conn, self._vmutils._SERIAL_PORT_SETTING_DATA_CLASS,
            element_instance_id=mock_vmsettings.InstanceID)

    def test_set_vm_serial_port_conn(self):
        self._lookup_vm()
        mock_com_1 = mock.Mock()
        mock_com_2 = mock.Mock()

        self._vmutils._get_vm_serial_ports = mock.Mock(
            return_value=[mock_com_1, mock_com_2])

        self._vmutils.set_vm_serial_port_connection(
            mock.sentinel.vm_name,
            port_number=1,
            pipe_path=mock.sentinel.pipe_path)

        self.assertEqual([mock.sentinel.pipe_path], mock_com_1.Connection)
        self._vmutils._jobutils.modify_virt_resource.assert_called_once_with(
            mock_com_1)

    def test_get_serial_port_conns(self):
        self._lookup_vm()

        mock_com_1 = mock.Mock()
        mock_com_1.Connection = []

        mock_com_2 = mock.Mock()
        mock_com_2.Connection = [mock.sentinel.pipe_path]

        self._vmutils._get_vm_serial_ports = mock.Mock(
            return_value=[mock_com_1, mock_com_2])

        ret_val = self._vmutils.get_vm_serial_port_connections(
            mock.sentinel.vm_name)
        expected_ret_val = [mock.sentinel.pipe_path]

        self.assertEqual(expected_ret_val, ret_val)

    def test_list_instance_notes(self):
        vs = mock.MagicMock()
        attrs = {'ElementName': 'fake_name',
                 'Notes': ['4f54fb69-d3a2-45b7-bb9b-b6e6b3d893b3']}
        vs.configure_mock(**attrs)
        vs2 = mock.MagicMock(ElementName='fake_name2', Notes=None)
        self._vmutils._conn.Msvm_VirtualSystemSettingData.return_value = [vs,
                                                                          vs2]
        response = self._vmutils.list_instance_notes()

        self.assertEqual([(attrs['ElementName'], attrs['Notes'])], response)
        self._vmutils._conn.Msvm_VirtualSystemSettingData.assert_called_with(
            ['ElementName', 'Notes'],
            VirtualSystemType=self._vmutils._VIRTUAL_SYSTEM_TYPE_REALIZED)

    def test_modify_virtual_system(self):
        mock_vs_man_svc = self._vmutils._vs_man_svc
        mock_vmsetting = mock.MagicMock()
        fake_job_path = 'fake job path'
        fake_ret_val = 'fake return value'

        mock_vs_man_svc.ModifySystemSettings.return_value = (fake_job_path,
                                                             fake_ret_val)

        self._vmutils._modify_virtual_system(vmsetting=mock_vmsetting)

        mock_vs_man_svc.ModifySystemSettings.assert_called_once_with(
            SystemSettings=mock_vmsetting.GetText_(1))
        self._vmutils._jobutils.check_ret_val.assert_called_once_with(
            fake_ret_val, fake_job_path)

    @ddt.data(True, False)
    @mock.patch.object(vmutils.VMUtils, '_get_wmi_obj')
    def test_create_vm(self, mock_get_wmi_obj, vnuma_enabled=True):
        mock_vs_man_svc = self._vmutils._vs_man_svc
        mock_vs_data = mock.MagicMock()
        fake_job_path = 'fake job path'
        fake_ret_val = 'fake return value'
        fake_vm_name = 'fake_vm_name'
        _conn = self._vmutils._conn.Msvm_VirtualSystemSettingData

        self._vmutils._jobutils.check_ret_val.return_value = mock.sentinel.job
        _conn.new.return_value = mock_vs_data
        mock_vs_man_svc.DefineSystem.return_value = (fake_job_path,
                                                     mock.sentinel.vm_path,
                                                     fake_ret_val)

        self._vmutils.create_vm(vm_name=fake_vm_name,
                                vm_gen=constants.VM_GEN_2,
                                notes='fake notes',
                                vnuma_enabled=vnuma_enabled,
                                instance_path=mock.sentinel.instance_path)

        _conn.new.assert_called_once_with()
        self.assertEqual(mock_vs_data.ElementName, fake_vm_name)
        mock_vs_man_svc.DefineSystem.assert_called_once_with(
            ResourceSettings=[], ReferenceConfiguration=None,
            SystemSettings=mock_vs_data.GetText_(1))
        self._vmutils._jobutils.check_ret_val.assert_called_once_with(
            fake_ret_val, fake_job_path)

        self.assertEqual(self._vmutils._VIRTUAL_SYSTEM_SUBTYPE_GEN2,
                         mock_vs_data.VirtualSystemSubType)
        self.assertFalse(mock_vs_data.SecureBootEnabled)

        self.assertEqual(vnuma_enabled, mock_vs_data.VirtualNumaEnabled)
        self.assertEqual(self._vmutils._VIRTUAL_SYSTEM_SUBTYPE_GEN2,
                         mock_vs_data.VirtualSystemSubType)
        self.assertEqual(mock_vs_data.Notes, 'fake notes')
        self.assertEqual(mock.sentinel.instance_path,
                         mock_vs_data.ConfigurationDataRoot)
        self.assertEqual(mock.sentinel.instance_path, mock_vs_data.LogDataRoot)
        self.assertEqual(mock.sentinel.instance_path,
                         mock_vs_data.SnapshotDataRoot)
        self.assertEqual(mock.sentinel.instance_path,
                         mock_vs_data.SuspendDataRoot)
        self.assertEqual(mock.sentinel.instance_path,
                         mock_vs_data.SwapFileDataRoot)

    def test_list_instances(self):
        vs = mock.MagicMock()
        attrs = {'ElementName': 'fake_name'}
        vs.configure_mock(**attrs)
        self._vmutils._conn.Msvm_VirtualSystemSettingData.return_value = [vs]
        response = self._vmutils.list_instances()

        self.assertEqual([(attrs['ElementName'])], response)
        self._vmutils._conn.Msvm_VirtualSystemSettingData.assert_called_with(
            ['ElementName'],
            VirtualSystemType=self._vmutils._VIRTUAL_SYSTEM_TYPE_REALIZED)

    def test_get_attached_disks(self):
        mock_scsi_ctrl_path = mock.MagicMock()
        expected_query = ("SELECT * FROM %(class_name)s "
                          "WHERE (ResourceSubType='%(res_sub_type)s' OR "
                          "ResourceSubType='%(res_sub_type_virt)s' OR "
                          "ResourceSubType='%(res_sub_type_dvd)s') AND "
                          "Parent = '%(parent)s'" %
                          {"class_name":
                           self._vmutils._RESOURCE_ALLOC_SETTING_DATA_CLASS,
                           "res_sub_type":
                           self._vmutils._PHYS_DISK_RES_SUB_TYPE,
                           "res_sub_type_virt":
                           self._vmutils._DISK_DRIVE_RES_SUB_TYPE,
                           "res_sub_type_dvd":
                           self._vmutils._DVD_DRIVE_RES_SUB_TYPE,
                           "parent": mock_scsi_ctrl_path.replace("'", "''")})
        expected_disks = self._vmutils._conn.query.return_value

        ret_disks = self._vmutils.get_attached_disks(mock_scsi_ctrl_path)

        self._vmutils._conn.query.assert_called_once_with(expected_query)
        self.assertEqual(expected_disks, ret_disks)

    def _get_fake_instance_notes(self):
        return [self._FAKE_VM_UUID]

    def test_instance_notes(self):
        mock_vm_settings = self._lookup_vm()
        mock_vm_settings.Notes = self._get_fake_instance_notes()

        notes = self._vmutils._get_instance_notes(mock.sentinel.vm_name)

        self.assertEqual(notes[0], self._FAKE_VM_UUID)

    def test_get_event_wql_query(self):
        cls = self._vmutils._COMPUTER_SYSTEM_CLASS
        field = self._vmutils._VM_ENABLED_STATE_PROP
        timeframe = 10
        filtered_states = [constants.HYPERV_VM_STATE_ENABLED,
                           constants.HYPERV_VM_STATE_DISABLED]

        expected_checks = ' OR '.join(
            ["TargetInstance.%s = '%s'" % (field, state)
             for state in filtered_states])
        expected_query = (
            "SELECT %(field)s, TargetInstance "
            "FROM __InstanceModificationEvent "
            "WITHIN %(timeframe)s "
            "WHERE TargetInstance ISA '%(class)s' "
            "AND TargetInstance.%(field)s != "
            "PreviousInstance.%(field)s "
            "AND (%(checks)s)"
            % {'class': cls,
               'field': field,
               'timeframe': timeframe,
               'checks': expected_checks})

        query = self._vmutils._get_event_wql_query(
            cls=cls, field=field, timeframe=timeframe,
            filtered_states=filtered_states)
        self.assertEqual(expected_query, query)

    def test_get_vm_power_state_change_listener(self):
        with mock.patch.object(self._vmutils,
                               '_get_event_wql_query') as mock_get_query:
            listener = self._vmutils.get_vm_power_state_change_listener(
                timeframe=mock.sentinel.timeframe,
                filtered_states=mock.sentinel.filtered_states)

            mock_get_query.assert_called_once_with(
                cls=self._vmutils._COMPUTER_SYSTEM_CLASS,
                field=self._vmutils._VM_ENABLED_STATE_PROP,
                timeframe=mock.sentinel.timeframe,
                filtered_states=mock.sentinel.filtered_states)
            watcher = self._vmutils._conn.Msvm_ComputerSystem.watch_for
            watcher.assert_called_once_with(
                raw_wql=mock_get_query.return_value,
                fields=[self._vmutils._VM_ENABLED_STATE_PROP])

            self.assertEqual(watcher.return_value, listener)

    @mock.patch('time.sleep')
    @mock.patch.object(vmutils, 'tpool')
    @mock.patch.object(vmutils, 'patcher')
    def test_vm_power_state_change_event_handler(self, mock_patcher,
                                                 mock_tpool, mock_sleep):
        enabled_state = constants.HYPERV_VM_STATE_ENABLED
        hv_enabled_state = self._vmutils._vm_power_states_map[enabled_state]
        fake_event = mock.Mock(ElementName=mock.sentinel.vm_name,
                               EnabledState=hv_enabled_state)
        fake_callback = mock.Mock(side_effect=Exception)

        fake_listener = (
            self._vmutils._conn.Msvm_ComputerSystem.watch_for.return_value)
        mock_tpool.execute.side_effect = (exceptions.x_wmi_timed_out,
                                          fake_event, Exception,
                                          KeyboardInterrupt)

        handler = self._vmutils.get_vm_power_state_change_listener(
            get_handler=True)
        # This is supposed to run as a daemon, so we'll just cause an
        # exception in order to be able to test the method.
        self.assertRaises(KeyboardInterrupt, handler, fake_callback)

        fake_callback.assert_called_once_with(mock.sentinel.vm_name,
                                              enabled_state)
        mock_tpool.execute.assert_has_calls(
            fake_listener,
            [mock.call(constants.DEFAULT_WMI_EVENT_TIMEOUT_MS)] * 4)
        mock_sleep.assert_called_once_with(
            constants.DEFAULT_WMI_EVENT_TIMEOUT_MS / 1000)

    def _test_get_vm_generation(self, vm_gen):
        mock_settings = self._lookup_vm()
        vm_gen_string = "Microsoft:Hyper-V:SubType:" + str(vm_gen)
        mock_settings.VirtualSystemSubType = vm_gen_string

        ret = self._vmutils.get_vm_generation(mock.sentinel.FAKE_VM_NAME)

        self.assertEqual(vm_gen, ret)

    def test_get_vm_generation_gen1(self):
        self._test_get_vm_generation(constants.VM_GEN_1)

    def test_get_vm_generation_gen2(self):
        self._test_get_vm_generation(constants.VM_GEN_2)

    def test_get_vm_generation_no_attr(self):
        mock_settings = self._lookup_vm()
        mock_settings.VirtualSystemSubType.side_effect = AttributeError

        ret = self._vmutils.get_vm_generation(mock.sentinel.FAKE_VM_NAME)

        self.assertEqual(constants.VM_GEN_1, ret)

    def test_stop_vm_jobs(self):
        mock_vm = self._lookup_vm()

        self._vmutils.stop_vm_jobs(mock.sentinel.vm_name)

        self._vmutils._jobutils.stop_jobs.assert_called_once_with(
            mock_vm, None)

    def test_set_secure_boot(self):
        vs_data = mock.MagicMock()
        self._vmutils._set_secure_boot(vs_data, msft_ca_required=False)
        self.assertTrue(vs_data.SecureBootEnabled)

    def test_set_secure_boot_CA_required(self):
        self.assertRaises(exceptions.HyperVException,
                          self._vmutils._set_secure_boot,
                          mock.MagicMock(), True)

    @mock.patch.object(vmutils.VMUtils, '_modify_virtual_system')
    @mock.patch.object(vmutils.VMUtils, '_lookup_vm_check')
    def test_enable_secure_boot(self, mock_lookup_vm_check,
                                mock_modify_virtual_system):
        vs_data = mock_lookup_vm_check.return_value

        with mock.patch.object(self._vmutils,
                               '_set_secure_boot') as mock_set_secure_boot:
            self._vmutils.enable_secure_boot(
                mock.sentinel.VM_NAME, mock.sentinel.certificate_required)

            mock_lookup_vm_check.assert_called_with(mock.sentinel.VM_NAME)
            mock_set_secure_boot.assert_called_once_with(
                vs_data, mock.sentinel.certificate_required)
            mock_modify_virtual_system.assert_called_once_with(vs_data)

    def test_set_disk_qos_specs_exc(self):
        self.assertRaises(exceptions.UnsupportedOperation,
                          self._vmutils.set_disk_qos_specs,
                          mock.sentinel.disk_path, mock.sentinel.max_iops)

    def test_set_disk_qos_specs_noop(self):
        self._vmutils.set_disk_qos_specs(mock.sentinel.disk_path, 0, 0)

    @ddt.data(
        {'drive_path':
            r'\\ADCONTROLLER\root\virtualization\v2:Msvm_DiskDrive.'
            r'CreationClassName="Msvm_DiskDrive",DeviceID="Microsoft:'
            r'6344C73D-6FD6-4A74-8BE8-8EEAC2737369\\0\\0\\D",'
            r'SystemCreationClassName="Msvm_ComputerSystem"',
         'exp_phys_disk': True},
        {'drive_path': 'some_image.vhdx',
         'exp_phys_disk': False})
    @ddt.unpack
    @mock.patch.object(vmutils.VMUtils,
                       '_get_mounted_disk_resource_from_path')
    def test_drive_to_boot_source(self, mock_get_disk_res_from_path,
                                  drive_path, exp_phys_disk):
        mock_drive = mock.MagicMock()
        mock_drive.Parent = mock.sentinel.bssd
        mock_get_disk_res_from_path.return_value = mock_drive

        exp_rasd_path = (mock_drive.path_.return_value
                         if exp_phys_disk else mock_drive.Parent)
        mock_same_element = mock.MagicMock()
        self._vmutils._conn.Msvm_LogicalIdentity.return_value = [
            mock.Mock(SameElement=mock_same_element)]

        ret = self._vmutils._drive_to_boot_source(drive_path)

        self._vmutils._conn.Msvm_LogicalIdentity.assert_called_once_with(
            SystemElement=exp_rasd_path)
        mock_get_disk_res_from_path.assert_called_once_with(
            drive_path, is_physical=exp_phys_disk)
        expected_path = mock_same_element.path_.return_value
        self.assertEqual(expected_path, ret)

    @mock.patch.object(vmutils.VMUtils, '_set_boot_order_gen1')
    @mock.patch.object(vmutils.VMUtils, '_set_boot_order_gen2')
    @mock.patch.object(vmutils.VMUtils, 'get_vm_generation')
    def _test_set_boot_order(self, mock_get_vm_gen, mock_set_boot_order_gen2,
                             mock_set_boot_order_gen1, vm_gen):
        mock_get_vm_gen.return_value = vm_gen
        self._vmutils.set_boot_order(mock.sentinel.fake_vm_name,
                                     mock.sentinel.boot_order)
        if vm_gen == constants.VM_GEN_1:
            mock_set_boot_order_gen1.assert_called_once_with(
                mock.sentinel.fake_vm_name, mock.sentinel.boot_order)
        else:
            mock_set_boot_order_gen2.assert_called_once_with(
                mock.sentinel.fake_vm_name, mock.sentinel.boot_order)

    def test_set_boot_order_gen1_vm(self):
        self._test_set_boot_order(vm_gen=constants.VM_GEN_1)

    def test_set_boot_order_gen2_vm(self):
        self._test_set_boot_order(vm_gen=constants.VM_GEN_2)

    @mock.patch.object(vmutils.VMUtils, '_modify_virtual_system')
    def test_set_boot_order_gen1(self, mock_modify_virt_syst):
        mock_vssd = self._lookup_vm()

        fake_dev_boot_order = [mock.sentinel.BOOT_DEV1,
                               mock.sentinel.BOOT_DEV2]
        self._vmutils._set_boot_order_gen1(
            mock_vssd.name, fake_dev_boot_order)

        mock_modify_virt_syst.assert_called_once_with(mock_vssd)
        self.assertEqual(mock_vssd.BootOrder, tuple(fake_dev_boot_order))

    @mock.patch.object(vmutils.VMUtils, '_drive_to_boot_source')
    @mock.patch.object(vmutils.VMUtils, '_modify_virtual_system')
    def test_set_boot_order_gen2(self, mock_modify_virtual_system,
                                 mock_drive_to_boot_source):
        fake_dev_order = ['fake_boot_source1', 'fake_boot_source2']
        mock_drive_to_boot_source.side_effect = fake_dev_order
        mock_vssd = self._lookup_vm()
        old_boot_order = tuple(['fake_boot_source2',
                                'fake_boot_source1',
                                'fake_boot_source_net'])
        expected_boot_order = tuple(['FAKE_BOOT_SOURCE1',
                                     'FAKE_BOOT_SOURCE2',
                                     'FAKE_BOOT_SOURCE_NET'])
        mock_vssd.BootSourceOrder = old_boot_order

        self._vmutils._set_boot_order_gen2(mock_vssd.name, fake_dev_order)

        mock_modify_virtual_system.assert_called_once_with(mock_vssd)
        self.assertEqual(expected_boot_order, mock_vssd.BootSourceOrder)

    def test_vm_gen_1_supports_remotefx(self):
        ret = self._vmutils.vm_gen_supports_remotefx(constants.VM_GEN_1)
        self.assertTrue(ret)

    def test_vm_gen_2_supports_remotefx(self):
        ret = self._vmutils.vm_gen_supports_remotefx(constants.VM_GEN_2)
        self.assertFalse(ret)

    def test_validate_remotefx_monitor_count(self):
        self.assertRaises(exceptions.HyperVRemoteFXException,
                          self._vmutils._validate_remotefx_params,
                          10, constants.REMOTEFX_MAX_RES_1024x768)

    def test_validate_remotefx_max_resolution(self):
        self.assertRaises(exceptions.HyperVRemoteFXException,
                          self._vmutils._validate_remotefx_params,
                          1, '1024x700')

    @ddt.data(True, False)
    @mock.patch.object(vmutils.VMUtils, '_set_remotefx_vram')
    @mock.patch.object(vmutils.VMUtils, '_get_new_resource_setting_data')
    def test_set_remotefx_display_controller(self, new_obj, mock_get_new_rsd,
                                             mock_set_remotefx_vram):
        if new_obj:
            remotefx_ctrl_res = None
            expected_res = mock_get_new_rsd.return_value
        else:
            remotefx_ctrl_res = mock.MagicMock()
            expected_res = remotefx_ctrl_res

        self._vmutils._set_remotefx_display_controller(
            mock.sentinel.fake_vm, remotefx_ctrl_res,
            mock.sentinel.monitor_count, mock.sentinel.max_resolution,
            mock.sentinel.vram_bytes)

        self.assertEqual(mock.sentinel.monitor_count,
                         expected_res.MaximumMonitors)
        self.assertEqual(mock.sentinel.max_resolution,
                         expected_res.MaximumScreenResolution)
        mock_set_remotefx_vram.assert_called_once_with(
            expected_res, mock.sentinel.vram_bytes)

        if new_obj:
            mock_get_new_rsd.assert_called_once_with(
                self._vmutils._REMOTEFX_DISP_CTRL_RES_SUB_TYPE,
                self._vmutils._REMOTEFX_DISP_ALLOCATION_SETTING_DATA_CLASS)
            self._vmutils._jobutils.add_virt_resource.assert_called_once_with(
                expected_res, mock.sentinel.fake_vm)
        else:
            self.assertFalse(mock_get_new_rsd.called)
            modify_virt_res = self._vmutils._jobutils.modify_virt_resource
            modify_virt_res.assert_called_once_with(expected_res)

    def test_set_remotefx_vram(self):
        self._vmutils._set_remotefx_vram(mock.sentinel.remotefx_ctrl_res,
                                         mock.sentinel.vram_bytes)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    @mock.patch.object(vmutils.VMUtils, '_set_remotefx_display_controller')
    @mock.patch.object(vmutils.VMUtils, '_vm_has_s3_controller')
    def test_enable_remotefx_video_adapter(self,
                                           mock_vm_has_s3_controller,
                                           mock_set_remotefx_ctrl,
                                           mock_get_element_associated_class):
        mock_vm = self._lookup_vm()

        mock_r1 = mock.MagicMock()
        mock_r1.ResourceSubType = self._vmutils._SYNTH_DISP_CTRL_RES_SUB_TYPE

        mock_r2 = mock.MagicMock()
        mock_r2.ResourceSubType = self._vmutils._S3_DISP_CTRL_RES_SUB_TYPE

        mock_get_element_associated_class.return_value = [mock_r1, mock_r2]

        self._vmutils.enable_remotefx_video_adapter(
            mock.sentinel.fake_vm_name,
            self._FAKE_MONITOR_COUNT,
            constants.REMOTEFX_MAX_RES_1024x768)

        mock_get_element_associated_class.assert_called_once_with(
            self._vmutils._conn,
            self._vmutils._CIM_RES_ALLOC_SETTING_DATA_CLASS,
            element_instance_id=mock_vm.InstanceID)
        self._vmutils._jobutils.remove_virt_resource.assert_called_once_with(
            mock_r1)

        mock_set_remotefx_ctrl.assert_called_once_with(
            mock_vm, None, self._FAKE_MONITOR_COUNT,
            self._vmutils._remote_fx_res_map[
                constants.REMOTEFX_MAX_RES_1024x768],
            None)

        self._vmutils._jobutils.modify_virt_resource.assert_called_once_with(
            mock_r2)
        self.assertEqual(self._vmutils._DISP_CTRL_ADDRESS_DX_11,
                         mock_r2.Address)

    @mock.patch.object(vmutils.VMUtils, '_vm_has_s3_controller')
    @mock.patch.object(vmutils.VMUtils, '_get_new_resource_setting_data')
    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_disable_remotefx_video_adapter(self,
                                            mock_get_element_associated_class,
                                            mock_get_new_rsd,
                                            mock_vm_has_s3_controller):
        mock_vm = self._lookup_vm()
        mock_r1 = mock.MagicMock(
            ResourceSubType=self._vmutils._REMOTEFX_DISP_CTRL_RES_SUB_TYPE)
        mock_r2 = mock.MagicMock(
            ResourceSubType=self._vmutils._S3_DISP_CTRL_RES_SUB_TYPE)

        mock_get_element_associated_class.return_value = [mock_r1, mock_r2]

        self._vmutils.disable_remotefx_video_adapter(
            mock.sentinel.fake_vm_name)

        mock_get_element_associated_class.assert_called_once_with(
            self._vmutils._conn,
            self._vmutils._CIM_RES_ALLOC_SETTING_DATA_CLASS,
            element_instance_id=mock_vm.InstanceID)
        self._vmutils._jobutils.remove_virt_resource.assert_called_once_with(
            mock_r1)
        mock_get_new_rsd.assert_called_once_with(
            self._vmutils._SYNTH_DISP_CTRL_RES_SUB_TYPE,
            self._vmutils._SYNTH_DISP_ALLOCATION_SETTING_DATA_CLASS)
        self._vmutils._jobutils.add_virt_resource.assert_called_once_with(
            mock_get_new_rsd.return_value, mock_vm)
        self._vmutils._jobutils.modify_virt_resource.assert_called_once_with(
            mock_r2)
        self.assertEqual(self._vmutils._DISP_CTRL_ADDRESS, mock_r2.Address)

    @mock.patch.object(_wqlutils, 'get_element_associated_class')
    def test_disable_remotefx_video_adapter_not_found(
            self, mock_get_element_associated_class):
        mock_vm = self._lookup_vm()
        mock_get_element_associated_class.return_value = []

        self._vmutils.disable_remotefx_video_adapter(
            mock.sentinel.fake_vm_name)

        mock_get_element_associated_class.assert_called_once_with(
            self._vmutils._conn,
            self._vmutils._CIM_RES_ALLOC_SETTING_DATA_CLASS,
            element_instance_id=mock_vm.InstanceID)
        self.assertFalse(self._vmutils._jobutils.remove_virt_resource.called)

    @mock.patch.object(vmutils.VMUtils, 'get_vm_generation')
    def test_vm_has_s3_controller(self, mock_get_vm_generation):
        self.assertTrue(self._vmutils._vm_has_s3_controller(
            mock.sentinel.fake_vm_name))

    @mock.patch.object(vmutils.VMUtils, '_get_mounted_disk_resource_from_path')
    def test_update_vm_disk_path(self, mock_get_disk_resource_from_path):
        disk_resource = mock_get_disk_resource_from_path.return_value
        self._vmutils.update_vm_disk_path(mock.sentinel.disk_path,
                                          mock.sentinel.new_path,
                                          is_physical=True)

        mock_get_disk_resource_from_path.assert_called_once_with(
            disk_path=mock.sentinel.disk_path, is_physical=True)
        self._vmutils._jobutils.modify_virt_resource.assert_called_once_with(
            disk_resource)
        self.assertEqual(disk_resource.HostResource, [mock.sentinel.new_path])

    def test_add_pci_device(self):
        self.assertRaises(NotImplementedError,
                          self._vmutils.add_pci_device,
                          mock.sentinel.vm_name, mock.sentinel.vendor_id,
                          mock.sentinel.product_id)

    def test_remove_pci_device(self):
        self.assertRaises(NotImplementedError,
                          self._vmutils.remove_pci_device,
                          mock.sentinel.vm_name, mock.sentinel.vendor_id,
                          mock.sentinel.product_id)

    def test_remove_all_pci_devices(self):
        self._vmutils.remove_all_pci_devices(mock.sentinel.vm_name)

    def test_populate_fsk(self):
        self.assertRaises(NotImplementedError,
                          self._vmutils.populate_fsk,
                          mock.sentinel.fsk_filepath,
                          mock.sentinel.fsk_pairs)

    def test_add_vtpm(self):
        self.assertRaises(NotImplementedError,
                          self._vmutils.add_vtpm,
                          mock.sentinel.vm_name, mock.sentinel.pdk_filepath,
                          mock.sentinel.shielded)

    def test_provision_vm(self):
        self.assertRaises(NotImplementedError,
                          self._vmutils.provision_vm,
                          mock.sentinel.vm_name, mock.sentinel.fsk_filepath,
                          mock.sentinel.pdk_filepath)


class VMUtils6_3TestCase(test_base.OsWinBaseTestCase):

    def setUp(self):
        super(VMUtils6_3TestCase, self).setUp()
        self._vmutils = vmutils.VMUtils6_3()
        self._vmutils._conn_attr = mock.MagicMock()
        self._vmutils._jobutils = mock.MagicMock()

    @mock.patch.object(vmutils.VMUtils,
                       '_get_mounted_disk_resource_from_path')
    def test_set_disk_qos_specs(self, mock_get_disk_resource):
        mock_disk = mock_get_disk_resource.return_value

        self._vmutils.set_disk_qos_specs(mock.sentinel.disk_path,
                                         max_iops=mock.sentinel.max_iops,
                                         min_iops=mock.sentinel.min_iops)

        mock_get_disk_resource.assert_called_once_with(
            mock.sentinel.disk_path, is_physical=False)
        self.assertEqual(mock.sentinel.max_iops, mock_disk.IOPSLimit)
        self.assertEqual(mock.sentinel.min_iops, mock_disk.IOPSReservation)
        self._vmutils._jobutils.modify_virt_resource.assert_called_once_with(
            mock_disk)

    @mock.patch.object(vmutils.VMUtils,
                       '_get_mounted_disk_resource_from_path')
    def test_set_disk_qos_specs_missing_values(self, mock_get_disk_resource):
        self._vmutils.set_disk_qos_specs(mock.sentinel.disk_path)

        self.assertFalse(mock_get_disk_resource.called)



# Copyright 2014 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes
import struct
import sys

from eventlet import patcher
from oslo_log import log as logging
from oslo_utils import units
import six

from os_win import _utils
from os_win import constants
from os_win import exceptions
from os_win.utils import win32utils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi import libs as w_lib
from os_win.utils.winapi import wintypes

kernel32 = w_lib.get_shared_lib_handle(w_lib.KERNEL32)

LOG = logging.getLogger(__name__)

# Avoid using six.moves.queue as we need a non monkey patched class
if sys.version_info > (3, 0):
    Queue = patcher.original('queue')
else:
    Queue = patcher.original('Queue')

WAIT_PIPE_DEFAULT_TIMEOUT = 5  # seconds
WAIT_IO_COMPLETION_TIMEOUT = 2 * units.k
WAIT_INFINITE_TIMEOUT = 0xFFFFFFFF

IO_QUEUE_TIMEOUT = 2
IO_QUEUE_BURST_TIMEOUT = 0.05


class IOUtils(object):
    """Asyncronous IO helper class."""

    def __init__(self):
        self._win32_utils = win32utils.Win32Utils()

    def _run_and_check_output(self, *args, **kwargs):
        eventlet_blocking_mode = kwargs.get('eventlet_nonblocking_mode', False)
        kwargs.update(kernel32_lib_func=True,
                      failure_exc=exceptions.Win32IOException,
                      eventlet_nonblocking_mode=eventlet_blocking_mode)
        return self._win32_utils.run_and_check_output(*args, **kwargs)

    @_utils.retry_decorator(exceptions=exceptions.Win32IOException,
                            max_sleep_time=2)
    def wait_named_pipe(self, pipe_name, timeout=WAIT_PIPE_DEFAULT_TIMEOUT):
        """Wait a given amount of time for a pipe to become available."""
        self._run_and_check_output(kernel32.WaitNamedPipeW,
                                   ctypes.c_wchar_p(pipe_name),
                                   timeout * units.k)

    def open(self, path, desired_access=None, share_mode=None,
             creation_disposition=None, flags_and_attributes=None):
        error_ret_vals = [w_const.INVALID_HANDLE_VALUE]
        handle = self._run_and_check_output(kernel32.CreateFileW,
                                            ctypes.c_wchar_p(path),
                                            desired_access,
                                            share_mode,
                                            None,
                                            creation_disposition,
                                            flags_and_attributes,
                                            None,
                                            error_ret_vals=error_ret_vals)
        return handle

    def close_handle(self, handle):
        self._run_and_check_output(kernel32.CloseHandle, handle)

    def cancel_io(self, handle, overlapped_structure=None,
                  ignore_invalid_handle=False):
        """Cancels pending IO on specified handle.

        If an overlapped structure is passed, only the IO requests that
        were issued with the specified overlapped structure are canceled.
        """
        # Ignore errors thrown when there are no requests
        # to be canceled.
        ignored_error_codes = [w_const.ERROR_NOT_FOUND]
        if ignore_invalid_handle:
            ignored_error_codes.append(w_const.ERROR_INVALID_HANDLE)
        lp_overlapped = (ctypes.byref(overlapped_structure)
                         if overlapped_structure else None)

        self._run_and_check_output(kernel32.CancelIoEx,
                                   handle,
                                   lp_overlapped,
                                   ignored_error_codes=ignored_error_codes)

    def _wait_io_completion(self, event):
        # In order to cancel this, we simply set the event.
        self._run_and_check_output(kernel32.WaitForSingleObjectEx,
                                   event, WAIT_INFINITE_TIMEOUT,
                                   True, error_ret_vals=[w_const.WAIT_FAILED])

    def set_event(self, event):
        self._run_and_check_output(kernel32.SetEvent, event)

    def _reset_event(self, event):
        self._run_and_check_output(kernel32.ResetEvent, event)

    def _create_event(self, event_attributes=None, manual_reset=True,
                      initial_state=False, name=None):
        return self._run_and_check_output(kernel32.CreateEventW,
                                          event_attributes, manual_reset,
                                          initial_state, name,
                                          error_ret_vals=[None])

    def get_completion_routine(self, callback=None):
        def _completion_routine(error_code, num_bytes, lpOverLapped):
            """Sets the completion event and executes callback, if passed."""
            overlapped = ctypes.cast(lpOverLapped,
                                     wintypes.LPOVERLAPPED).contents
            self.set_event(overlapped.hEvent)

            if callback:
                callback(num_bytes)

        return wintypes.LPOVERLAPPED_COMPLETION_ROUTINE(_completion_routine)

    def get_new_overlapped_structure(self):
        """Structure used for asynchronous IO operations."""
        # Event used for signaling IO completion
        hEvent = self._create_event()

        overlapped_structure = wintypes.OVERLAPPED()
        overlapped_structure.hEvent = hEvent
        return overlapped_structure

    def read(self, handle, buff, num_bytes,
             overlapped_structure, completion_routine):
        self._reset_event(overlapped_structure.hEvent)
        self._run_and_check_output(kernel32.ReadFileEx,
                                   handle, buff, num_bytes,
                                   ctypes.byref(overlapped_structure),
                                   completion_routine)
        self._wait_io_completion(overlapped_structure.hEvent)

    def write(self, handle, buff, num_bytes,
              overlapped_structure, completion_routine):
        self._reset_event(overlapped_structure.hEvent)
        self._run_and_check_output(kernel32.WriteFileEx,
                                   handle, buff, num_bytes,
                                   ctypes.byref(overlapped_structure),
                                   completion_routine)
        self._wait_io_completion(overlapped_structure.hEvent)

    @classmethod
    def get_buffer(cls, buff_size, data=None):
        buff = (ctypes.c_ubyte * buff_size)()
        if data:
            cls.write_buffer_data(buff, data)
        return buff

    @staticmethod
    def get_buffer_data(buff, num_bytes):
        return bytes(bytearray(buff[:num_bytes]))

    @staticmethod
    def write_buffer_data(buff, data):
        for i, c in enumerate(data):
            buff[i] = struct.unpack('B', six.b(c))[0]


class IOQueue(Queue.Queue, object):
    def __init__(self, client_connected):
        Queue.Queue.__init__(self)
        self._client_connected = client_connected

    def get(self, timeout=IO_QUEUE_TIMEOUT, continue_on_timeout=True):
        while self._client_connected.isSet():
            try:
                return Queue.Queue.get(self, timeout=timeout)
            except Queue.Empty:
                if continue_on_timeout:
                    continue
                else:
                    break

    def put(self, item, timeout=IO_QUEUE_TIMEOUT):
        while self._client_connected.isSet():
            try:
                return Queue.Queue.put(self, item, timeout=timeout)
            except Queue.Full:
                continue

    def get_burst(self, timeout=IO_QUEUE_TIMEOUT,
                  burst_timeout=IO_QUEUE_BURST_TIMEOUT,
                  max_size=constants.SERIAL_CONSOLE_BUFFER_SIZE):
        # Get as much data as possible from the queue
        # to avoid sending small chunks.
        data = self.get(timeout=timeout)

        while data and len(data) <= max_size:
            chunk = self.get(timeout=burst_timeout,
                             continue_on_timeout=False)
            if chunk:
                data += chunk
            else:
                break
        return data
